#!/usr/bin/env python
# License: BSD 3 clause
"""
Runs a bunch of scikit-learn jobs in parallel on the cluster given a
config file.

:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:author: Nitin Madnani (nmadnani@ets.org)
"""


import logging
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser

from skll.experiments import run_configuration
from skll.version import __version__


def main(argv=None):
    """
    Handles command line arguments and gets things started.

    Parameters
    ----------
    argv : list of str
        List of arguments, as if specified on the command-line.
        If None, ``sys.argv[1:]`` is used instead.
    """

    # Get command line arguments
    parser = ArgumentParser(description='Runs the scikit-learn experiments '
                                        'in a given config file. If Grid Map '
                                        'is installed, jobs will automatically'
                                        ' be created and run on a '
                                        'DRMAA-compatible cluster.',
                            formatter_class=ArgumentDefaultsHelpFormatter,
                            conflict_handler='resolve')
    parser.add_argument('config_file',
                        help='Configuration file describing the task to run.',
                        nargs='+')
    parser.add_argument('-a',
                        '--ablation',
                        help='Runs an ablation study where repeated '
                             'experiments are conducted where the specified '
                             'number of features in each featureset in the '
                             'configuration file are held out.',
                        type=int,
                        default=0,
                        metavar='NUM_FEATURES')
    parser.add_argument('-A',
                        '--ablation_all',
                        help='Runs an ablation study where repeated '
                             'experiments are conducted with all combinations '
                             'of features in each featureset in the '
                             'configuration file. Overrides --ablation '
                             'setting.',
                        action='store_true')
    parser.add_argument('-k',
                        '--keep_models',
                        help='If trained models already exists, re-use them '
                             'instead of overwriting them.',
                        action='store_true')
    parser.add_argument('-l',
                        '--local',
                        help='Do not use the Grid Engine for running jobs and'
                             ' just run everything sequentially on the local '
                             'machine. ',
                        action='store_true')
    parser.add_argument('-m',
                        '--machines',
                        help='comma-separated list of machines to add to the '
                             'gridmap whitelist (if not specified, all '
                             'available machines are used). Note that full '
                             'names must be specified, e.g., '
                             '"nlp.research.ets.org"',
                        default=None)
    parser.add_argument('-q',
                        '--queue',
                        help="Use this queue for gridmap.",
                        default='all.q')
    parser.add_argument('-r',
                        '--resume',
                        help='If result files already exist for an experiment'
                             ' do not overwrite them. This is very useful '
                             'when doing a large ablation experiment and part'
                             ' of it crashes.',
                        action='store_true')
    parser.add_argument('-v',
                        '--verbose',
                        help='Include debug information in the logging output.',
                        default=False,
                        action='store_true')
    parser.add_argument('--version',
                        action='version',
                        version=f'%(prog)s {__version__}')
    args = parser.parse_args(argv)

    # Default logging level is INFO unless we are being verbose
    log_level = logging.DEBUG if args.verbose else logging.INFO

    # Make warnings from built-in warnings module get formatted more nicely
    logging.captureWarnings(True)
    logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - '
                               '%(message)s',
                        level=log_level)

    machines = None
    if args.machines:
        machines = args.machines.split(',')

    ablation = args.ablation
    if args.ablation_all:
        ablation = None

    # Run each config file sequentially
    for config_file in args.config_file:
        run_configuration(config_file,
                          local=args.local,
                          overwrite=not args.keep_models,
                          queue=args.queue,
                          hosts=machines,
                          ablation=ablation,
                          resume=args.resume,
                          log_level=log_level)


if __name__ == '__main__':
    main()



# License: BSD 3 clause
"""
Metrics that can be used to evaluate the performance of learners.

:author: Nitin Madnani (nmadnani@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:author: Dan Blanchard (dblanchard@ets.org)
:organization: ETS
"""

import sys
from importlib import import_module
from inspect import signature
from os.path import abspath, basename, dirname, exists

import numpy as np
from scipy.stats import kendalltau, pearsonr, spearmanr
from sklearn.metrics import SCORERS, confusion_matrix, f1_score, make_scorer

# a set that will hold the names of any custom metrics;
# this is a private variable only meant for internal use
_CUSTOM_METRICS = set()


def kappa(y_true, y_pred, weights=None, allow_off_by_one=False):
    """
    Calculates the kappa inter-rater agreement between two the gold standard
    and the predicted ratings. Potential values range from -1 (representing
    complete disagreement) to 1 (representing complete agreement).  A kappa
    value of 0 is expected if all agreement is due to chance.

    In the course of calculating kappa, all items in ``y_true`` and ``y_pred`` will
    first be converted to floats and then rounded to integers.

    It is assumed that y_true and y_pred contain the complete range of possible
    ratings.

    This function contains a combination of code from yorchopolis's kappa-stats
    and Ben Hamner's Metrics projects on Github.

    Parameters
    ----------
    y_true : array-like of float
        The true/actual/gold labels for the data.
    y_pred : array-like of float
        The predicted/observed labels for the data.
    weights : str or np.array, optional
        Specifies the weight matrix for the calculation.
        Options are ::

            -  None = unweighted-kappa
            -  'quadratic' = quadratic-weighted kappa
            -  'linear' = linear-weighted kappa
            -  two-dimensional numpy array = a custom matrix of

        weights. Each weight corresponds to the
        :math:`w_{ij}` values in the wikipedia description
        of how to calculate weighted Cohen's kappa.
        Defaults to None.
    allow_off_by_one : bool, optional
        If true, ratings that are off by one are counted as
        equal, and all other differences are reduced by
        one. For example, 1 and 2 will be considered to be
        equal, whereas 1 and 3 will have a difference of 1
        for when building the weights matrix.
        Defaults to False.

    Returns
    -------
    k : float
        The kappa score, or weighted kappa score.

    Raises
    ------
    AssertionError
        If ``y_true`` != ``y_pred``.
    ValueError
        If labels cannot be converted to int.
    ValueError
        If invalid weight scheme.
    """

    # Ensure that the lists are both the same length
    assert(len(y_true) == len(y_pred))

    # This rather crazy looking typecast is intended to work as follows:
    # If an input is an int, the operations will have no effect.
    # If it is a float, it will be rounded and then converted to an int
    # because the ml_metrics package requires ints.
    # If it is a str like "1", then it will be converted to a (rounded) int.
    # If it is a str that can't be typecast, then the user is
    # given a hopefully useful error message.
    try:
        y_true = [int(np.round(float(y))) for y in y_true]
        y_pred = [int(np.round(float(y))) for y in y_pred]
    except ValueError:
        raise ValueError("For kappa, the labels should be integers or strings"
                         " that can be converted to ints (E.g., '4.0' or "
                         "'3').")

    # Figure out normalized expected values
    min_rating = min(min(y_true), min(y_pred))
    max_rating = max(max(y_true), max(y_pred))

    # shift the values so that the lowest value is 0
    # (to support scales that include negative values)
    y_true = [y - min_rating for y in y_true]
    y_pred = [y - min_rating for y in y_pred]

    # Build the observed/confusion matrix
    num_ratings = max_rating - min_rating + 1
    observed = confusion_matrix(y_true, y_pred,
                                labels=list(range(num_ratings)))
    num_scored_items = float(len(y_true))

    # Build weight array if weren't passed one
    if isinstance(weights, str):
        wt_scheme = weights
        weights = None
    else:
        wt_scheme = ''
    if weights is None:
        weights = np.empty((num_ratings, num_ratings))
        for i in range(num_ratings):
            for j in range(num_ratings):
                diff = abs(i - j)
                if allow_off_by_one and diff:
                    diff -= 1
                if wt_scheme == 'linear':
                    weights[i, j] = diff
                elif wt_scheme == 'quadratic':
                    weights[i, j] = diff ** 2
                elif not wt_scheme:  # unweighted
                    weights[i, j] = bool(diff)
                else:
                    raise ValueError('Invalid weight scheme specified for '
                                     f'kappa: {wt_scheme}')

    hist_true = np.bincount(y_true, minlength=num_ratings)
    hist_true = hist_true[: num_ratings] / num_scored_items
    hist_pred = np.bincount(y_pred, minlength=num_ratings)
    hist_pred = hist_pred[: num_ratings] / num_scored_items
    expected = np.outer(hist_true, hist_pred)

    # Normalize observed array
    observed = observed / num_scored_items

    # If all weights are zero, that means no disagreements matter.
    k = 1.0
    if np.count_nonzero(weights):
        k -= (sum(sum(weights * observed)) / sum(sum(weights * expected)))

    return k


def correlation(y_true, y_pred, corr_type='pearson'):
    """
    Calculate given correlation between ``y_true`` and ``y_pred``. ``y_pred``
    can be multi-dimensional. If ``y_pred`` is 1-dimensional, it may either
    contain probabilities, most-likely classification labels, or regressor
    predictions. In that case, we simply return the correlation between
    ``y_true`` and ``y_pred``. If ``y_pred`` is multi-dimensional,
    it contains probabilties for multiple classes in which case, we infer
    the most likely labels and then compute the correlation between those
    and ``y_true``.

    Parameters
    ----------
    y_true : array-like of float
        The true/actual/gold labels for the data.
    y_pred : array-like of float
        The predicted/observed labels for the data.
    corr_type : str, optional
        Which type of correlation to compute. Possible
        choices are ``pearson``, ``spearman``,
        and ``kendall_tau``.
        Defaults to ``pearson``.

    Returns
    -------
    ret_score : float
        correlation value if well-defined, else 0.0
    """

    # get the correlation function to use based on the given type
    corr_func = pearsonr
    if corr_type == 'spearman':
        corr_func = spearmanr
    elif corr_type == 'kendall_tau':
        corr_func = kendalltau

    # convert to numpy array in case we are passed a list
    y_pred = np.array(y_pred)

    # multi-dimensional -> probability array -> get label
    if y_pred.ndim > 1:
        labels = np.argmax(y_pred, axis=1)
        ret_score = corr_func(y_true, labels)[0]
    # 1-dimensional -> probabilities/labels -> use as is
    else:
        ret_score = corr_func(y_true, y_pred)[0]
    return ret_score


def f1_score_least_frequent(y_true, y_pred):
    """
    Calculate the F1 score of the least frequent label/class in ``y_true`` for
    ``y_pred``.

    Parameters
    ----------
    y_true : array-like of float
        The true/actual/gold labels for the data.
    y_pred : array-like of float
        The predicted/observed labels for the data.

    Returns
    -------
    ret_score : float
        F1 score of the least frequent label.
    """
    least_frequent = np.bincount(y_true).argmin()
    return f1_score(y_true, y_pred, average=None)[least_frequent]


def register_custom_metric(custom_metric_path, custom_metric_name):
    """
    Import, load, and register the custom metric function from the given path.

    Parameters
    ----------
    custom_metric_path : str
        The path to a custom metric.
    custom_metric_name : str
        The name of the custom metric function to load.
        This function must take only two array-like
        arguments: the true labels and the predictions,
        in that order.

    Raises
    ------
    ValueError
        If the custom metric path does not end in '.py'.
    NameError
        If the name of the custom metric file conflicts
        with an already existing attribute in ``skll.metrics``
        or if the custom metric name conflicts with a scikit-learn
        or SKLL metric.
    """
    if not custom_metric_path:
        raise ValueError(f"custom metric path was not set and "
                         f"metric {custom_metric_name} was not found.")

    if not exists(custom_metric_path):
        raise ValueError(f"custom metric path '{custom_metric_path}' "
                         f"does not exist.")

    if not custom_metric_path.endswith('.py'):
        raise ValueError(f"custom metric path must end in .py, you specified "
                         f"{custom_metric_path}")

    # get the name of the module containing the custom metric
    custom_metric_module_name = basename(custom_metric_path)[:-3]

    # once we know that the module name is okay, we need to make sure
    # that the metric function name is also okay
    if custom_metric_name in SCORERS:
        raise NameError(f"a metric called '{custom_metric_name}' already "
                        f"exists in SKLL; rename the metric function "
                        f"in {custom_metric_module_name}.py and try again.")

    # dynamically import the module unless we have already done it
    if custom_metric_module_name not in sys.modules:
        sys.path.append(dirname(abspath(custom_metric_path)))
        metric_module = import_module(custom_metric_module_name)

        # this statement is only necessary so that if we end
        # up using multiprocessing parallelization backend,
        # things are serialized properly
        globals()[custom_metric_module_name] = metric_module

    # get the metric function from this imported module
    metric_func = getattr(sys.modules[custom_metric_module_name],
                          custom_metric_name)
    # again, we need this for multiprocessing serialization
    metric_func.__module__ = f"skll.metrics.{custom_metric_module_name}"

    # extract any "special" keyword arguments from the metric function
    metric_func_parameters = signature(metric_func).parameters
    make_scorer_kwargs = {}
    for make_scorer_kwarg in ['greater_is_better',
                              'needs_proba',
                              'needs_threshold']:
        if make_scorer_kwarg in metric_func_parameters:
            parameter_value = metric_func_parameters.get(make_scorer_kwarg).default
            make_scorer_kwargs.update({make_scorer_kwarg: parameter_value})

    # make the scorer function with the extracted keyword arguments
    # and add it to the `CUSTOM_METRICS` set
    SCORERS[f"{custom_metric_name}"] = make_scorer(metric_func, **make_scorer_kwargs)
    _CUSTOM_METRICS.add(custom_metric_name)

    return metric_func


def use_score_func(func_name, y_true, y_pred):
    """
    Call the scoring function in ``sklearn.metrics.SCORERS`` with the given
    name. This takes care of handling keyword arguments that were pre-specified
    when creating the scorer. This applies any sign-flipping that was specified
    by ``make_scorer()`` when the scorer was created.

    Parameters
    ----------
    func_name : str
        The name of the objective function to use from SCORERS.
    y_true : array-like of float
        The true/actual/gold labels for the data.
    y_pred : array-like of float
        The predicted/observed labels for the data.

    Returns
    -------
    ret_score : float
        The scored result from the given scorer.
    """
    scorer = SCORERS[func_name]
    return scorer._sign * scorer._score_func(y_true, y_pred, **scorer._kwargs)



#!/usr/bin/env python
# License: BSD 3 clause
"""
script for computing additional evaluation metrics

:author: Michael Heilman (mheilman@ets.org)
"""

import argparse
import csv
import logging

from numpy.random import RandomState

from skll.data import Reader, safe_float
from skll.metrics import use_score_func
from skll.version import __version__

# Make warnings from built-in warnings module get formatted more nicely
logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - '
                           '%(message)s')
logger = logging.getLogger(__name__)


def get_prediction_from_probabilities(classes,
                                      probabilities,
                                      prediction_method,
                                      random_state=1234567890):
    """
    Convert a list of class-probabilities into a class prediction. This function assumes
    that, if the prediction method is 'expected_value', the class labels are integers.

    Parameters
    ----------
    classes: list of str or int
        List of string or integer class names.
    probabilities: list of float
        List of probabilities for the respective classes.
    prediction_method: str
        Indicates how to get a single class prediction from the probabilities.
        Must be one of:
            1. "highest": Selects the class with the highest probability. If
               multiple classes have the same probability, a class is selected randomly.
            2. "expected_value": Calculates an expected value over integer classes and
               rounds to the nearest int.

    random_state: int
        Seed for `RandomState`, used for randomly selecting a class when necessary.

    Returns
    -------
    predicted_class: str or int
        Predicted class.

    """
    prng = RandomState(random_state)
    if prediction_method == 'highest':
        highest_p = max(probabilities)
        best_classes = [classes[i] for i, p in enumerate(probabilities) if p == highest_p]
        if len(best_classes) > 1:
            return prng.choice(best_classes)
        else:
            return best_classes[0]

    elif prediction_method == 'expected_value':
        exp_val = sum([classes[i] * prob for i, prob in enumerate(probabilities)])
        return int(round(exp_val))


def compute_eval_from_predictions(examples_file,
                                  predictions_file,
                                  metric_names,
                                  prediction_method=None):
    """
    Compute evaluation metrics from prediction files after you have run an
    experiment.

    Parameters
    ----------
    examples_file: str
        Path to a SKLL examples file (in .jsonlines or other format).
    predictions_file: str
        Path to a SKLL predictions output TSV file with id and prediction column names.
    metric_names: list of str
        A list of SKLL metric names (e.g., [pearson, unweighted_kappa]).
    prediction_method: str or None
        Indicates how to get a single class prediction from the probabilities. Currently
        supported options are  "highest", which selects the class with the highest
        probability, and "expected_value", which calculates an expected value over
        integer classes and rounds to the nearest int. If predictions file does not
        contain probabilities, this should be set to None.

    Returns
    -------
    dict
        Maps metrics names to corresponding values.

    Raises
    ------
    ValueError
        If the requested prediction method is 'expected_value' but the class names can't
        be converted to ints.
    """

    # read gold standard labels
    data = Reader.for_path(examples_file).read()
    gold = dict(zip(data.ids, data.labels))

    # read predictions
    pred = {}
    with open(predictions_file) as pred_file:
        reader = csv.reader(pred_file, dialect=csv.excel_tab)
        header = next(reader)

        # If there are more than two columns, assume column 0 contains the ids, and
        # columns 1-n contain class probabilities. Convert them to a class prediction
        # using the specified `method`.
        if len(header) > 2:
            classes = [c for c in header[1:] if c]
            if prediction_method is None:
                prediction_method = "highest"
                logger.info("No prediction method specified. Using 'highest'.")
            if prediction_method == 'expected_value':
                try:
                    classes = [int(c) for c in classes]
                except ValueError as e:
                    raise e
            for row in reader:
                probabilities = [safe_float(p) for p in row[1:]]
                prediction = get_prediction_from_probabilities(classes,
                                                               probabilities,
                                                               prediction_method)
                pred[row[0]] = safe_float(prediction)
        else:
            if prediction_method is not None:
                logger.warning("A prediction method was provided, but the "
                               "predictions file doesn't contain "
                               "probabilities. Ignoring prediction method "
                               f"'{prediction_method}'.")

            for row in reader:
                pred[row[0]] = safe_float(row[1])

    # make a sorted list of example ids in order to match up
    # labels and predictions
    if set(gold.keys()) != set(pred.keys()):
        raise ValueError('The example and prediction IDs do not match.')
    example_ids = sorted(gold.keys())

    res = {}
    for metric_name in metric_names:
        score = use_score_func(metric_name,
                               [gold[ex_id] for ex_id in example_ids],
                               [pred[ex_id] for ex_id in example_ids])
        res[metric_name] = score
    return res


def main(argv=None):
    """
    Handles command line arguments and gets things started.

    Parameters
    ----------
    argv: list of str
        List of arguments, as if specified on the command-line. If None, `sys.argv[1:]`
        is used instead.
    """
    # Get command line arguments
    parser = argparse.ArgumentParser(
        description="Computes evaluation metrics from prediction files after "
                    "you have run an experiment.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument('examples_file',
                        help='SKLL input file with labeled examples')
    parser.add_argument('predictions_file',
                        help='file with predictions from SKLL')
    parser.add_argument('metric_names',
                        help='metrics to compute',
                        nargs='+')
    parser.add_argument('--method', '-m',
                        help="How to generate a prediction from the class "
                             "probabilities. Supported methods are 'highest' "
                             "(default) and 'expected_value' (only works with"
                             " integer classes).")
    parser.add_argument('--version', action='version',
                        version=f'%(prog)s {__version__}')
    args = parser.parse_args(argv)

    supported_prediction_methods = {"highest", "expected_value"}
    if (args.method is not None) and (args.method not in supported_prediction_methods):
        raise KeyError(f"Unrecognized prediction method '{args.method}'. "
                       "Supported methods are 'highest' and "
                       "'expected_value'.")

    scores = compute_eval_from_predictions(args.examples_file,
                                           args.predictions_file,
                                           args.metric_names,
                                           args.method)

    for metric_name in args.metric_names:
        print(f"{scores[metric_name]}\t{metric_name}\t{args.predictions_file}")


if __name__ == '__main__':
    main()



# License: BSD 3 clause
"""
This package provides a number of utilities to make it simpler to run
common scikit-learn experiments with pre-generated features.

:author: Nitin Madnani (nmadnani@ets.org)
:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:organization: ETS
"""

from sklearn.metrics import SCORERS, f1_score, fbeta_score, make_scorer

from .metrics import correlation, f1_score_least_frequent, kappa

# Add our scorers to the sklearn dictionary here so that they will always be
# available if you import anything from skll
_scorers = {"f1_score_micro": make_scorer(f1_score, average="micro"),
            "f1_score_macro": make_scorer(f1_score, average="macro"),
            "f1_score_weighted": make_scorer(f1_score, average="weighted"),
            "f1_score_least_frequent": make_scorer(f1_score_least_frequent),
            "f05": make_scorer(fbeta_score, beta=0.5, average="binary"),
            "f05_score_micro": make_scorer(fbeta_score, beta=0.5, average="micro"),
            "f05_score_macro": make_scorer(fbeta_score, beta=0.5, average="macro"),
            "f05_score_weighted": make_scorer(fbeta_score, beta=0.5, average="weighted"),
            "pearson": make_scorer(correlation, corr_type="pearson"),
            "spearman": make_scorer(correlation, corr_type="spearman"),
            "kendall_tau": make_scorer(correlation, corr_type="kendall_tau"),
            "unweighted_kappa": make_scorer(kappa),
            "quadratic_weighted_kappa": make_scorer(kappa, weights="quadratic"),
            "linear_weighted_kappa": make_scorer(kappa, weights="linear"),
            "qwk_off_by_one": make_scorer(kappa,
                                          weights="quadratic",
                                          allow_off_by_one=True),
            "lwk_off_by_one": make_scorer(kappa,
                                          weights="linear",
                                          allow_off_by_one=True),
            "uwk_off_by_one": make_scorer(kappa, allow_off_by_one=True)}

SCORERS.update(_scorers)



#!/usr/bin/env python
# License: BSD 3 clause
"""
Script that converts feature files from one format to another

:author: Nitin Madnani (nmadnani@ets.org)
:organization: ETS
"""

import argparse
import logging
import os
import sys

from bs4 import UnicodeDammit

from skll.data.dict_vectorizer import DictVectorizer
from skll.data.readers import EXT_TO_READER
from skll.data.writers import (
    EXT_TO_WRITER,
    ARFFWriter,
    CSVWriter,
    LibSVMWriter,
    TSVWriter,
)
from skll.version import __version__


def _pair_to_dict_tuple(pair):
    """
    A helper function for constructing mappings from feature/class names to
    numbers.
    """
    number, name = pair.split('=')
    number = int(number)
    return (name, number)


def main(argv=None):
    """
    Handles command line arguments and gets things started.

    Parameters
    ----------
    argv : list of str
        List of arguments, as if specified on the command-line.
        If None, ``sys.argv[1:]`` is used instead.
    """

    # Get command line arguments
    parser = argparse.ArgumentParser(
        description="Takes an input feature file and converts it to another "
                    "format. Formats are determined automatically from file "
                    "extensions.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument('infile',
                        help='input feature file (ends in .arff, .csv, '
                             '.jsonlines, .libsvm, .ndj, or .tsv)')
    parser.add_argument('outfile',
                        help='output feature file (ends in .arff, .csv, '
                             '.jsonlines, .libsvm, .ndj, or .tsv)')
    parser.add_argument('-i', '--id_col',
                        help='Name of the column which contains the instance '
                             'IDs in ARFF, CSV, or TSV files.',
                        default='id')
    label_group = parser.add_mutually_exclusive_group(required=False)
    label_group.add_argument('-l',
                             '--label_col',
                             help='Name of the column which contains the '
                                  'class labels in ARFF, CSV, or TSV files. '
                                  'For ARFF files, this must be the final '
                                  'column to count as the label.',
                             default='y')
    label_group.add_argument('--no_labels',
                             action='store_true',
                             default=False,
                             help='Used to indicate that the input data has '
                                  'no labels.')
    parser.add_argument('-q', '--quiet',
                        help='Suppress printing of "Loading..." messages.',
                        action='store_true')
    parser.add_argument('--arff_regression',
                        help='Create ARFF files for regression, not '
                             'classification.',
                        action='store_true')
    parser.add_argument('--arff_relation',
                        help='Relation name to use for ARFF file.',
                        default='skll_relation')
    parser.add_argument('--reuse_libsvm_map',
                        help='If you want to output multiple files that use '
                             'the same mapping from labels and features to '
                             'numbers when writing libsvm files, you can '
                             'specify an existing .libsvm file to reuse the '
                             'mapping from.',
                        type=argparse.FileType('rb'))
    parser.add_argument('--version', action='version',
                        version=f'%(prog)s {__version__}')
    args = parser.parse_args(argv)

    # Make warnings from built-in warnings module get formatted more nicely
    logging.captureWarnings(True)
    logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - '
                               '%(message)s')
    logger = logging.getLogger(__name__)

    # make sure the input file extension is one we can process
    input_extension = os.path.splitext(args.infile)[1].lower()
    output_extension = os.path.splitext(args.outfile)[1].lower()

    if input_extension not in EXT_TO_READER:
        logger.error('Input file must be in either .arff, .csv, .jsonlines, '
                     '.libsvm, .ndj, or .tsv format. You specified: '
                     f'{input_extension}')
        sys.exit(1)

    # Build feature and label vectorizers from existing libsvm file if asked
    if args.reuse_libsvm_map and output_extension == '.libsvm':
        feat_map = {}
        label_map = {}
        for line in args.reuse_libsvm_map:
            line = UnicodeDammit(line, ['utf-8',
                                        'windows-1252']).unicode_markup
            if '#' not in line:
                logger.error('The LibSVM file you want to reuse the map from '
                             'was not created by SKLL and does not actually '
                             'contain the necessary mapping info.')
                sys.exit(1)
            comments = line.split('#')[1]
            _, label_map_str, feat_map_str = comments.split('|')
            feat_map.update(_pair_to_dict_tuple(pair) for pair in
                            feat_map_str.strip().split())
            label_map.update(_pair_to_dict_tuple(pair) for pair in
                             label_map_str
                             .strip().split())
        feat_vectorizer = DictVectorizer()
        feat_vectorizer.fit([{name: 1} for name in feat_map])
        feat_vectorizer.vocabulary_ = feat_map
    else:
        feat_vectorizer = None
        label_map = None

    label_col = None if args.no_labels else args.label_col

    # Iterate through input file and collect the information we need
    reader = EXT_TO_READER[input_extension](args.infile,
                                            quiet=args.quiet,
                                            label_col=label_col,
                                            id_col=args.id_col)
    feature_set = reader.read()
    # write out the file in the requested output format
    writer_type = EXT_TO_WRITER[output_extension]
    writer_args = {'quiet': args.quiet}
    if writer_type is CSVWriter or writer_type is TSVWriter:
        writer_args['label_col'] = label_col
        writer_args['id_col'] = args.id_col
    elif writer_type is ARFFWriter:
        writer_args['label_col'] = label_col
        writer_args['id_col'] = args.id_col
        writer_args['regression'] = args.arff_regression
        writer_args['relation'] = args.arff_relation
    elif writer_type is LibSVMWriter:
        writer_args['label_map'] = label_map
    writer = writer_type(args.outfile, feature_set, **writer_args)
    writer.write()


if __name__ == '__main__':
    main()



# License: BSD 3 clause
"""
Functions related to logging in SKLL.

:author: Nitin Madnani (nmadnani@ets.org)
:organization: ETS
"""
import logging
import re
import warnings
from functools import partial
from os.path import sep

orig_showwarning = warnings.showwarning
SKLEARN_WARNINGS_RE = re.compile(re.escape(f"{sep}sklearn{sep}"))


def send_sklearn_warnings_to_logger(logger, message, category, filename,
                                    lineno, file=None, line=None):
    """
    Return method that sends `sklearn`-specific warnings to a logger
    that can be used to replace warnings.showwarning (via `partial`,
    specifying a `logger` instance).
    """

    if SKLEARN_WARNINGS_RE.search(filename):
        logger.warning(f'{filename}:{lineno}: {category.__name__}:{message}')
    else:
        orig_showwarning(message, category, filename, lineno,
                         file=file, line=line)


def get_skll_logger(name, filepath=None, log_level=logging.INFO):
    """
    Create and return logger instances that are appropriate for use
    in SKLL code, e.g., that they can log to both STDERR as well
    as a file. This function will try to reuse any previously created
    logger based on the given name and filepath.

    Parameters
    ----------
    name : str
        The name to be used for the logger.
    filepath : str, optional
        The file to be used for the logger via a FileHandler.
        Default: None in which case no file is attached to the
        logger.
        Defaults to ``None``.
    log_level : str, optional
        The level for logging messages
        Defaults to ``logging.INFO``.

    Returns
    -------
    logger: logging.Logger
        A ``Logger`` instance.
    """

    # first get the logger instance associated with the
    # given name if one already exists
    logger = logging.getLogger(name)
    logger.setLevel(log_level)

    # if we are given a file path and this existing logger doesn't already
    # have a file handler for this file, then add one.
    if filepath:
        def is_file_handler(handler):
            return isinstance(handler, logging.FileHandler) and handler.stream.name == filepath
        need_file_handler = not any([is_file_handler(handler) for handler in logger.handlers])
        if need_file_handler:
            formatter = logging.Formatter('%(asctime)s - %(levelname)s - '
                                          '%(message)s')
            file_handler = logging.FileHandler(filepath, mode='w')
            file_handler.setFormatter(formatter)
            file_handler.setLevel(log_level)
            logger.addHandler(file_handler)

    warnings.showwarning = partial(send_sklearn_warnings_to_logger, logger)

    # return the logger instance
    return logger


def close_and_remove_logger_handlers(logger):
    """
    Close and remove any handlers attached to a logger instance.

    Parameters
    ----------
    logger : logging.Logger
        Logger instance

    Returns
    -------
    NoneType
    """

    for handler in logger.handlers[:]:
        handler.close()
        logger.removeHandler(handler)



"""
Constants useful for SKLL learners.

:author: Nitin Madnani (nmadnani@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:author: Dan Blanchard (dblanchard@ets.org)
:author: Aoife Cahill (acahill@ets.org)
:organization: ETS
"""

import os

from sklearn.dummy import DummyClassifier, DummyRegressor
from sklearn.ensemble import (
    AdaBoostClassifier,
    AdaBoostRegressor,
    GradientBoostingClassifier,
    GradientBoostingRegressor,
    RandomForestClassifier,
    RandomForestRegressor,
)
from sklearn.linear_model import (
    BayesianRidge,
    ElasticNet,
    HuberRegressor,
    Lars,
    Lasso,
    LinearRegression,
    LogisticRegression,
    RANSACRegressor,
    Ridge,
    RidgeClassifier,
    SGDClassifier,
    SGDRegressor,
    TheilSenRegressor,
)
from sklearn.naive_bayes import MultinomialNB
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from sklearn.neural_network import MLPClassifier, MLPRegressor
from sklearn.svm import SVC, SVR, LinearSVC, LinearSVR
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

KNOWN_DEFAULT_PARAM_GRIDS = {AdaBoostClassifier:
                             {'learning_rate': [0.01, 0.1, 1.0, 10.0, 100.0]},
                             AdaBoostRegressor:
                             {'learning_rate': [0.01, 0.1, 1.0, 10.0, 100.0]},
                             BayesianRidge:
                             {'alpha_1': [1e-6, 1e-4, 1e-2, 1, 10],
                              'alpha_2': [1e-6, 1e-4, 1e-2, 1, 10],
                              'lambda_1': [1e-6, 1e-4, 1e-2, 1, 10],
                              'lambda_2': [1e-6, 1e-4, 1e-2, 1, 10]},
                             DecisionTreeClassifier:
                             {'max_features': ["auto", None]},
                             DecisionTreeRegressor:
                             {'max_features': ["auto", None]},
                             DummyClassifier:
                             {},
                             DummyRegressor:
                             {},
                             ElasticNet:
                             {'alpha': [0.01, 0.1, 1.0, 10.0, 100.0]},
                             GradientBoostingClassifier:
                             {'max_depth': [1, 3, 5]},
                             GradientBoostingRegressor:
                             {'max_depth': [1, 3, 5]},
                             HuberRegressor:
                             {'epsilon': [1.05, 1.35, 1.5, 2.0, 2.5, 5.0],
                              'alpha': [1e-4, 1e-3, 1e-2, 1e-1, 1, 10, 100, 1000]},
                             KNeighborsClassifier:
                             {'n_neighbors': [1, 5, 10, 100],
                              'weights': ['uniform', 'distance']},
                             KNeighborsRegressor:
                             {'n_neighbors': [1, 5, 10, 100],
                              'weights': ['uniform', 'distance']},
                             MLPClassifier:
                             {'activation': ['logistic', 'tanh', 'relu'],
                              'alpha': [1e-4, 1e-3, 1e-2, 1e-1, 1],
                              'learning_rate_init': [0.001, 0.01, 0.1]},
                             MLPRegressor:
                             {'activation': ['logistic', 'tanh', 'relu'],
                              'alpha': [1e-4, 1e-3, 1e-2, 1e-1, 1],
                              'learning_rate_init': [0.001, 0.01, 0.1]},
                             MultinomialNB:
                             {'alpha': [0.1, 0.25, 0.5, 0.75, 1.0]},
                             Lars:
                             {},
                             Lasso:
                             {'alpha': [0.01, 0.1, 1.0, 10.0, 100.0]},
                             LinearRegression:
                             {},
                             LinearSVC:
                             {'C': [0.01, 0.1, 1.0, 10.0, 100.0]},
                             LogisticRegression:
                             {'C': [0.01, 0.1, 1.0, 10.0, 100.0]},
                             SVC:
                             {'C': [0.01, 0.1, 1.0, 10.0, 100.0],
                              'gamma': ['auto', 'scale', 0.01, 0.1, 1.0, 10.0, 100.0]},
                             RandomForestClassifier:
                             {'max_depth': [1, 5, 10, None]},
                             RandomForestRegressor:
                             {'max_depth': [1, 5, 10, None]},
                             RANSACRegressor:
                             {},
                             Ridge:
                             {'alpha': [0.01, 0.1, 1.0, 10.0, 100.0]},
                             RidgeClassifier:
                             {'alpha': [0.01, 0.1, 1.0, 10.0, 100.0]},
                             SGDClassifier:
                             {'alpha': [0.000001, 0.00001, 0.0001, 0.001, 0.01],
                              'penalty': ['l1', 'l2', 'elasticnet']},
                             SGDRegressor:
                             {'alpha': [0.000001, 0.00001, 0.0001, 0.001, 0.01],
                              'penalty': ['l1', 'l2', 'elasticnet']},
                             LinearSVR:
                             {'C': [0.01, 0.1, 1.0, 10.0, 100.0]},
                             SVR:
                             {'C': [0.01, 0.1, 1.0, 10.0, 100.0],
                              'gamma': ['auto', 'scale', 0.01, 0.1, 1.0, 10.0, 100.0]},
                             TheilSenRegressor:
                             {}
                             }

KNOWN_REQUIRES_DENSE = (BayesianRidge, Lars, TheilSenRegressor)

MAX_CONCURRENT_PROCESSES = int(os.getenv('SKLL_MAX_CONCURRENT_PROCESSES', '5'))

VALID_FEATURE_SCALING_OPTIONS = frozenset(['both',
                                           'none',
                                           'with_std',
                                           'with_mean'])

VALID_SAMPLERS = frozenset(['Nystroem',
                            'RBFSampler',
                            'SkewedChi2Sampler',
                            'AdditiveChi2Sampler',
                            ''])

VALID_TASKS = frozenset(['cross_validate',
                         'evaluate',
                         'learning_curve',
                         'predict',
                         'train'])

#: Set of evaluation metrics only used for classification tasks
CLASSIFICATION_ONLY_METRICS = {
    'accuracy',
    'average_precision',
    'balanced_accuracy',
    'f1',
    'f1_score_least_frequent',
    'f1_score_macro',
    'f1_score_micro',
    'f1_score_weighted',
    'f05',
    'f05_score_macro',
    'f05_score_micro',
    'f05_score_weighted',
    'jaccard',
    'jaccard_macro',
    'jaccard_micro',
    'jaccard_weighted',
    'neg_log_loss',
    'precision',
    'precision_macro',
    'precision_micro',
    'precision_weighted',
    'recall',
    'recall_macro',
    'recall_micro',
    'recall_weighted',
    'roc_auc'
}


#: Set of evaluation metrics based on correlation
CORRELATION_METRICS = {'kendall_tau', 'pearson', 'spearman'}

#: Set of evaluation metrics that can use prediction probabilities
PROBABILISTIC_METRICS = frozenset(['average_precision',
                                   'neg_log_loss',
                                   'roc_auc'])

#: Set of evaluation metrics only used for regression tasks
REGRESSION_ONLY_METRICS = {
    'explained_variance',
    'max_error',
    'neg_mean_squared_error',
    'neg_mean_absolute_error',
    'r2'
}

#: Set of unweighted kappa agreement metrics
UNWEIGHTED_KAPPA_METRICS = {'unweighted_kappa', 'uwk_off_by_one'}

#: Set of weighed kappa agreement metrics
WEIGHTED_KAPPA_METRICS = {
    'linear_weighted_kappa',
    'lwk_off_by_one',
    'quadratic_weighted_kappa',
    'qwk_off_by_one'
}



# License: BSD 3 clause
"""
Utility classes and functions for running SKLL experiments.

:author: Nitin Madnani (nmadnani@ets.org)
:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
"""

import json
import math
import re
from collections import defaultdict

import numpy as np
from sklearn.pipeline import Pipeline
from tabulate import tabulate

from skll.utils.logging import get_skll_logger


class NumpyTypeEncoder(json.JSONEncoder):
    """
    This class is used when serializing results, particularly the input label
    values if the input has int-valued labels.  Numpy int64 objects can't
    be serialized by the json module, so we must convert them to int objects.

    A related issue where this was adapted from:
    https://stackoverflow.com/questions/11561932/why-does-json-dumpslistnp-arange5-fail-while-json-dumpsnp-arange5-tolis
    """

    def default(self, obj):
        if isinstance(obj, (np.int32, np.int64)):
            return int(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class PipelineTypeEncoder(json.JSONEncoder):
    """
    This class is used for serializing ``sklearn.pipeline.Pipeline`` objects.
    """

    def default(self, obj):
        if isinstance(obj, Pipeline):
            pipeline_steps = str(obj.named_steps)
            pipeline_steps = re.sub(r'\n', '', pipeline_steps)
            pipeline_steps = re.sub(r'\s+', ' ', pipeline_steps)
            return pipeline_steps
        return json.JSONEncoder.default(self, obj)


def _check_job_results(job_results):
    """
    See if we have a complete results dictionary for every job.

    Parameters
    ----------
    job_results : list of dicts
        A list of job result dictionaries.
    """
    logger = get_skll_logger('experiment')
    logger.info('Checking job results')
    for result_dicts in job_results:
        if not result_dicts or 'task' not in result_dicts[0]:
            logger.error('There was an error running the experiment:\n'
                         f'{result_dicts}')


def _create_learner_result_dicts(task_results,
                                 grid_scores,
                                 grid_search_cv_results_dicts,
                                 learner_result_dict_base):
    """
    Create the learner result dictionaries that are used to create JSON and
    plain-text results files.

    Parameters
    ----------
    task_results : list
        The task results list.
    grid_scores : list
        The grid scores list.
    grid_search_cv_results_dicts : list of dicts
        A list of dictionaries of grid search CV results, one per fold,
        with keys such as "params", "mean_test_score", etc, that are
        mapped to lists of values associated with each hyperparameter set
        combination.
    learner_result_dict_base : dict
        Base dictionary for all learner results.

    Returns
    -------
    res : list of dicts
        The results of the learners, as a list of
        dictionaries.
    """
    res = []

    num_folds = len(task_results)
    accuracy_sum = 0.0
    pearson_sum = 0.0
    additional_metric_score_sums = {}
    score_sum = None
    prec_sum_dict = defaultdict(float)
    recall_sum_dict = defaultdict(float)
    f_sum_dict = defaultdict(float)
    result_table = None

    for (k,
         ((conf_matrix,
           fold_accuracy,
           result_dict,
           model_params,
           score,
           additional_scores),
          grid_score,
          grid_search_cv_results)) in enumerate(zip(task_results,
                                                    grid_scores,
                                                    grid_search_cv_results_dicts),
                                                start=1):

        # create a new dict for this fold
        learner_result_dict = {}
        learner_result_dict.update(learner_result_dict_base)

        # initialize some variables to blanks so that the
        # set of columns is fixed.
        learner_result_dict['result_table'] = ''
        learner_result_dict['accuracy'] = ''
        learner_result_dict['pearson'] = ''
        learner_result_dict['score'] = ''
        learner_result_dict['fold'] = ''

        if learner_result_dict_base['task'] == 'cross_validate':
            learner_result_dict['fold'] = k

        # include model parameters dump for regular learners only;
        # we need to use a special JSON encoder for voting learners
        # that contain ``Pipeline`` objects
        learner_result_dict['model_params'] = json.dumps(model_params,
                                                         cls=PipelineTypeEncoder)
        if grid_score is not None:
            learner_result_dict['grid_score'] = grid_score
            learner_result_dict['grid_search_cv_results'] = grid_search_cv_results

        if conf_matrix:
            labels = sorted(task_results[0][2].keys())
            headers = [""] + labels + ["Precision", "Recall", "F-measure"]
            rows = []
            for i, actual_label in enumerate(labels):
                conf_matrix[i][i] = f"[{conf_matrix[i][i]}]"
                label_prec = _get_stat_float(result_dict[actual_label],
                                             "Precision")
                label_recall = _get_stat_float(result_dict[actual_label],
                                               "Recall")
                label_f = _get_stat_float(result_dict[actual_label],
                                          "F-measure")
                if not math.isnan(label_prec):
                    prec_sum_dict[actual_label] += float(label_prec)
                if not math.isnan(label_recall):
                    recall_sum_dict[actual_label] += float(label_recall)
                if not math.isnan(label_f):
                    f_sum_dict[actual_label] += float(label_f)
                result_row = ([actual_label] + conf_matrix[i] +
                              [label_prec, label_recall, label_f])
                rows.append(result_row)

            result_table = tabulate(rows,
                                    headers=headers,
                                    stralign="right",
                                    floatfmt=".3f",
                                    tablefmt="grid")
            result_table_str = (f'{result_table}\n(row = reference; column = '
                                'predicted)')
            learner_result_dict['result_table'] = result_table_str
            learner_result_dict['accuracy'] = fold_accuracy
            accuracy_sum += fold_accuracy

        # if there is no confusion matrix, then we must be dealing
        # with a regression model
        else:
            learner_result_dict.update(result_dict)
            pearson_sum += float(learner_result_dict['pearson'])

        # get the scores for all the metrics and compute the sums
        if score is not None:
            if score_sum is None:
                score_sum = score
            else:
                score_sum += score
            learner_result_dict['score'] = score
        learner_result_dict['additional_scores'] = additional_scores
        for metric, score in additional_scores.items():
            if score is not None:
                additional_metric_score_sums[metric] = \
                    additional_metric_score_sums.get(metric, 0) + score
        res.append(learner_result_dict)

    if num_folds > 1:
        learner_result_dict = {}
        learner_result_dict.update(learner_result_dict_base)

        learner_result_dict['fold'] = 'average'

        if result_table:
            headers = ["Label", "Precision", "Recall", "F-measure"]
            rows = []
            for actual_label in labels:
                # Convert sums to means
                prec_mean = prec_sum_dict[actual_label] / num_folds
                recall_mean = recall_sum_dict[actual_label] / num_folds
                f_mean = f_sum_dict[actual_label] / num_folds
                rows.append([actual_label] + [prec_mean, recall_mean, f_mean])

            result_table = tabulate(rows,
                                    headers=headers,
                                    floatfmt=".3f",
                                    tablefmt="psql")
            learner_result_dict['result_table'] = str(result_table)
            learner_result_dict['accuracy'] = accuracy_sum / num_folds
        else:
            learner_result_dict['pearson'] = pearson_sum / num_folds

        if score_sum is not None:
            learner_result_dict['score'] = score_sum / num_folds
        scoredict = {}
        for metric, score_sum in additional_metric_score_sums.items():
            scoredict[metric] = score_sum / num_folds
        learner_result_dict['additional_scores'] = scoredict
        res.append(learner_result_dict)
    return res


def _get_stat_float(label_result_dict, stat):
    """
    A helper function to get output for the precision, recall, and f-score
    columns in the confusion matrix.

    Parameters
    ----------
    label_result_dict : dict
        Dictionary containing the stat we'd like
        to retrieve for a particular label.
    stat : str
        The statistic we're looking for in the dictionary.

    Returns
    -------
    stat_float : float
        The value of the stat if it's in the dictionary, and NaN
        otherwise.
    """
    if stat in label_result_dict and label_result_dict[stat] is not None:
        return label_result_dict[stat]
    else:
        return float('nan')



#!/usr/bin/env python
# License: BSD 3 clause
"""
A Helper script to generate learning plots from the learning curve output TSV
file.

This is necessary in scenarios where the plots were not generated as part of
the original learning curve experiment, e.g. the experiment was run on a remote
server where plots may not have been generated either due to a crash or
incorrect setting of the DISPLAY environment variable.

In these cases, the summary file should always be generated and this script can
then be used to generate the plots later.

:author: Nitin Madnani
:organization: ETS
"""

import argparse
import logging
import sys
from os import makedirs
from os.path import basename, exists

from skll.experiments import generate_learning_curve_plots
from skll.version import __version__


def main(argv=None):
    """
    Handles command line arguments and gets things started.

    Parameters
    ----------
    argv : list of str
        List of arguments, as if specified on the command-line.
        If None, ``sys.argv[1:]`` is used instead.
    """

    # Get command line arguments
    parser = argparse.ArgumentParser(
        description="Generates learning curve plots from the learning curve "
                    "TSV file.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        conflict_handler='resolve')
    parser.add_argument('tsv_file',
                        help='Learning curve TSV output file.')
    parser.add_argument('output_dir',
                        help='Directory to store the learning curve plots.')
    parser.add_argument('--version', action='version',
                        version=f'%(prog)s {__version__}')
    args = parser.parse_args(argv)

    # Make warnings from built-in warnings module get formatted more nicely
    logging.captureWarnings(True)
    logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - '
                               '%(message)s')

    # make sure that the input TSV file that's being passed exists
    if not exists(args.tsv_file):
        logging.error(f"Error: the given file {args.tsv_file} does not "
                      "exist.")
        sys.exit(1)

    # create the output directory if it doesn't already exist
    if not exists(args.output_dir):
        makedirs(args.output_dir)

    # get the experiment name from the learning curve TSV file
    # output_file_name = experiment_name + '_summary.tsv'
    experiment_name = basename(args.tsv_file).rstrip('_summary.tsv')
    logging.info("Generating learning curve(s)")
    generate_learning_curve_plots(experiment_name, args.output_dir, args.tsv_file)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# License: BSD 3 clause
"""
Simple script for printing out model weights.

:author: Michael Heilman (mheilman@ets.org)
:author: Nitin Madnani (nmadnani@ets.org)
:organization: ETS
"""

import argparse
import logging
import sys
from collections import defaultdict

import numpy as np
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC, LinearSVC

from skll.learner import Learner
from skll.version import __version__


def main(argv=None):
    """
    Handles command line arguments and gets things started.

    Parameters
    ----------
    argv : list of str
        List of arguments, as if specified on the command-line.
        If None, ``sys.argv[1:]`` is used instead.
    """

    parser = argparse.ArgumentParser(description="Prints out the weights of a"
                                                 " given model.",
                                     conflict_handler='resolve',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('model_file', help='model file to load')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--k',
                       help='number of top features to print (0 for all)',
                       type=int, default=50)
    group.add_argument("--sort_by_labels", '-s', action='store_true',
                       default=False, help="order the features by classes")
    parser.add_argument('--sign',
                        choices=['positive', 'negative', 'all'],
                        default='all',
                        help='show only positive, only negative or all weights')
    parser.add_argument('--version', action='version',
                        version=f'%(prog)s {__version__}')
    args = parser.parse_args(argv)

    # Make warnings from built-in warnings module get formatted more nicely
    logging.captureWarnings(True)
    logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - '
                               '%(message)s')

    k = args.k if args.k > 0 else None

    learner = Learner.from_file(args.model_file)
    (weights, intercept) = learner.model_params

    multiclass = False
    model = learner._model
    if (isinstance(model, LinearSVC)
        or (isinstance(model, LogisticRegression)
            and len(learner.label_list) > 2)
        or (isinstance(model, SVC)
            and model.kernel == 'linear')):
        multiclass = True
    weight_items = weights.items()
    if args.sign == 'positive':
        weight_items = (x for x in weight_items if x[1] > 0)
    elif args.sign == 'negative':
        weight_items = (x for x in weight_items if x[1] < 0)

    if intercept is not None:
        # subclass of LinearModel
        if '_intercept_' in intercept:
            # Some learners (e.g. LinearSVR) may return an array of intercepts but
            # sometimes that array is of length 1 so we don't need to print that
            # as an array/list. First, let's normalize these cases.
            model_intercepts = intercept['_intercept_']
            intercept_is_array = isinstance(model_intercepts, np.ndarray)
            num_intercepts = len(model_intercepts) if intercept_is_array else 1
            if intercept_is_array and num_intercepts == 1:
                model_intercepts = model_intercepts[0]
                intercept_is_array = False

            # now print out the intercepts
            print(f"intercept = {model_intercepts:.12f}")
        else:
            print("== intercept values ==")
            for (label, val) in intercept.items():
                print(f"{val:.12f}\t{label}")
        print()

    print("Number of nonzero features:", len(weights), file=sys.stderr)
    weight_by_class = defaultdict(dict)
    if multiclass and args.sort_by_labels:
        for label_feature, weight in weight_items:
            label, feature = label_feature.split()
            weight_by_class[label][feature] = weight
        for label in sorted(weight_by_class):
            for feat, val in sorted(weight_by_class[label].items(), key=lambda x: -abs(x[1])):
                print(f"{val:.12f}\t{label}\t{feat}")
    else:
        for feat, val in sorted(weight_items, key=lambda x: -abs(x[1]))[:k]:
            print(f"{val:.12f}\t{feat}")


if __name__ == '__main__':
    main()



# License: BSD 3 clause
"""
Functions for reading inputs for SKLL experiments.

:author: Nitin Madnani (nmadnani@ets.org)
:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
"""

from os.path import isfile, join

from skll.data.readers import Reader


def load_featureset(dir_path,
                    feat_files,
                    suffix,
                    id_col='id',
                    label_col='y',
                    ids_to_floats=False,
                    quiet=False,
                    class_map=None,
                    feature_hasher=False,
                    num_features=None,
                    logger=None):
    """
    Load a list of feature files and merge them.

    Parameters
    ----------
    dir_path : str
        Path to the directory that contains the feature files.
    feat_files : list of str
        A list of feature file prefixes.
    suffix : str
        The suffix to add to feature file prefixes to get the full filenames.
    id_col : str, optional
        Name of the column which contains the instance IDs.
        If no column with that name exists, or `None` is
        specified, example IDs will be automatically generated.
        Defaults to ``'id'``.
    label_col : str, optional
        Name of the column which contains the class labels.
        If no column with that name exists, or `None` is
        specified, the data is considered to be unlabeled.
        Defaults to ``'y'``.
    ids_to_floats : bool, optional
        Whether to convert the IDs to floats to save memory. Will raise error
        if we encounter non-numeric IDs.
        Defaults to ``False``.
    quiet : bool, optional
        Do not print "Loading..." status message to stderr.
        Defaults to ``False``.
    class_map : dict, optional
        Mapping from original class labels to new ones. This is
        mainly used for collapsing multiple labels into a single
        class. Anything not in the mapping will be kept the same.
        Defaults to ``None``.
    feature_hasher : bool, optional
        Should we use a FeatureHasher when vectorizing
        features?
        Defaults to ``False``.
    num_features : int, optional
        The number of features to use with the ``FeatureHasher``.
        This should always be set to the power of 2 greater
        than the actual number of features you're using.
        Defaults to ``None``.
    logger : logging.Logger, optional
        A logger instance to use to log messages instead of creating
        a new one by default.
        Defaults to ``None``.

    Returns
    -------
    merged_set : skll.FeatureSet
        A ``FeatureSet`` instance containing the specified labels, IDs, features,
        and feature vectorizer.
    """
    # if the training file is specified via train_file, then dir_path
    # actually contains the entire file name
    if isfile(dir_path):
        return Reader.for_path(dir_path,
                               label_col=label_col,
                               id_col=id_col,
                               ids_to_floats=ids_to_floats,
                               quiet=quiet,
                               class_map=class_map,
                               feature_hasher=feature_hasher,
                               num_features=num_features,
                               logger=logger).read()
    else:
        if len(feat_files) > 1 and feature_hasher:
            logger.warning("Since there are multiple feature files, "
                           "feature hashing applies to each specified "
                           "feature file separately.")
        merged_set = None
        for file_name in sorted(join(dir_path, featfile + suffix) for
                                featfile in feat_files):
            fs = Reader.for_path(file_name,
                                 label_col=label_col,
                                 id_col=id_col,
                                 ids_to_floats=ids_to_floats,
                                 quiet=quiet,
                                 class_map=class_map,
                                 feature_hasher=feature_hasher,
                                 num_features=num_features,
                                 logger=logger).read()
            if merged_set is None:
                merged_set = fs
            else:
                merged_set += fs
        return merged_set



# License: BSD 3 clause
"""
Functions for running and interacting with SKLL experiments.

:author: Nitin Madnani (nmadnani@ets.org)
:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:author: Chee Wee Leong (cleong@ets.org)
"""

import datetime
import json
import logging
from itertools import combinations
from os.path import exists, getsize, join

import matplotlib.pyplot as plt
import numpy as np
from sklearn import __version__ as SCIKIT_VERSION
from sklearn.metrics import SCORERS

from skll.config import parse_config_file
from skll.config.utils import _munge_featureset_name
from skll.learner import MAX_CONCURRENT_PROCESSES, Learner, load_custom_learner
from skll.learner.voting import VotingLearner
from skll.metrics import _CUSTOM_METRICS, register_custom_metric
from skll.utils.logging import close_and_remove_logger_handlers, get_skll_logger
from skll.version import __version__

from .input import load_featureset
from .output import (
    _print_fancy_output,
    _write_learning_curve_file,
    _write_skll_folds,
    _write_summary_file,
    generate_learning_curve_plots,
)
from .utils import NumpyTypeEncoder, _check_job_results, _create_learner_result_dicts

# Check if gridmap is available
try:
    from gridmap import Job, process_jobs
except ImportError:
    _HAVE_GRIDMAP = False
else:
    _HAVE_GRIDMAP = True

# Turn off interactive plotting for matplotlib
plt.ioff()


__all__ = ['generate_learning_curve_plots',
           'load_featureset',
           'run_configuration']


def _classify_featureset(args):  # noqa: C901
    """
    Classification job to be submitted to grid.

    Parameters
    ----------
    args : dict
        A dictionary with arguments for classifying the
        ``FeatureSet`` instance.

    Returns
    -------
    res : list of dicts
        The results of the classification, in the format
        of a list of dictionaries.

    Raises
    ------
    ValueError
        If extra unknown arguments are passed to the function.
    """

    # Extract all the arguments.
    # (There doesn't seem to be a better way to do this since one can't specify
    # required keyword arguments.)

    experiment_name = args.pop("experiment_name")
    task = args.pop("task")
    sampler = args.pop("sampler")
    feature_hasher = args.pop("feature_hasher")
    hasher_features = args.pop("hasher_features")
    job_name = args.pop("job_name")
    featureset = args.pop("featureset")
    featureset_name = args.pop("featureset_name")
    learner_name = args.pop("learner_name")
    train_path = args.pop("train_path")
    test_path = args.pop("test_path")
    train_set_name = args.pop("train_set_name")
    test_set_name = args.pop("test_set_name")
    shuffle = args.pop('shuffle')
    model_path = args.pop("model_path")
    prediction_prefix = args.pop("prediction_prefix")
    grid_search = args.pop("grid_search")
    grid_objective = args.pop("grid_objective")
    output_metrics = args.pop("output_metrics")
    suffix = args.pop("suffix")
    job_log_file = args.pop("log_file")
    job_log_level = args.pop("log_level")
    probability = args.pop("probability")
    pipeline = args.pop("pipeline")
    results_path = args.pop("results_path")
    fixed_parameters = args.pop("fixed_parameters")
    sampler_parameters = args.pop("sampler_parameters")
    param_grid = args.pop("param_grid")
    pos_label = args.pop("pos_label")
    overwrite = args.pop("overwrite")
    feature_scaling = args.pop("feature_scaling")
    min_feature_count = args.pop("min_feature_count")
    folds_file = args.pop("folds_file")
    grid_search_jobs = args.pop("grid_search_jobs")
    grid_search_folds = args.pop("grid_search_folds")
    cv_folds = args.pop("cv_folds")
    cv_seed = args.pop("cv_seed")
    save_cv_folds = args.pop("save_cv_folds")
    save_cv_models = args.pop("save_cv_models")
    use_folds_file_for_grid_search = args.pop("use_folds_file_for_grid_search")
    stratified_folds = args.pop("do_stratified_folds")
    label_col = args.pop("label_col")
    id_col = args.pop("id_col")
    ids_to_floats = args.pop("ids_to_floats")
    class_map = args.pop("class_map")
    custom_learner_path = args.pop("custom_learner_path")
    custom_metric_path = args.pop("custom_metric_path")
    quiet = args.pop('quiet', False)
    learning_curve_cv_folds = args.pop("learning_curve_cv_folds")
    learning_curve_train_sizes = args.pop("learning_curve_train_sizes")
    save_votes = args.pop("save_votes")

    if args:
        raise ValueError("Extra arguments passed to _classify_featureset: "
                         f"{args.keys()}")
    start_timestamp = datetime.datetime.now()

    # create a new SKLL logger for this specific job and
    # use the given log level
    logger = get_skll_logger(job_name,
                             job_log_file,
                             log_level=job_log_level)

    try:

        # log messages
        logger.info(f"Task: {task}")

        # check if we have any possible custom metrics
        possible_custom_metric_names = []
        for metric_name in output_metrics + [grid_objective]:
            # metrics that are not in `SCORERS` or `None` are candidates
            # (the `None` is a by-product of how jobs with single tuning
            # objectives are created)
            if metric_name not in SCORERS and metric_name is not None:
                possible_custom_metric_names.append(metric_name)
            # if the metric is already in `SCORERS`, is it a custom one
            # that we already registered? if so, log that
            elif metric_name in _CUSTOM_METRICS:
                logger.info(f"custom metric '{metric_name}' is already registered")

        # initialize list that will hold any invalid metrics
        # that we could not register as custom metrics
        invalid_metric_names = []

        # if we have possible custom metrics
        if possible_custom_metric_names:

            # check that we have a file to load them from
            if not custom_metric_path:
                raise ValueError(f"invalid metrics specified: {possible_custom_metric_names}")
            else:
                # try to register each possible custom metric
                # raise an exception if we fail, if we don't then
                # add the custom metric function to `globals()` so
                # that it serializes properly for gridmap
                for custom_metric_name in possible_custom_metric_names:
                    try:
                        custom_metric_func = register_custom_metric(custom_metric_path, custom_metric_name)
                    except (AttributeError, NameError, ValueError):
                        invalid_metric_names.append(custom_metric_name)
                    else:
                        logger.info(f"registered '{custom_metric_name}' as a "
                                    f"custom metric")
                        globals()[custom_metric_name] = custom_metric_func

        # raise an error if we have any invalid metrics
        if invalid_metric_names:
            raise ValueError(f"invalid metrics specified: {invalid_metric_names}. "
                             f"If these are custom metrics, check the function "
                             f"names.")

        if task == 'cross_validate':
            if isinstance(cv_folds, int):
                num_folds = cv_folds
            else:  # folds_file was used, so count the unique fold ids.
                num_folds = len(set(cv_folds.values()))
            logger.info(f"Cross-validating ({num_folds} folds, seed={cv_seed}) on "
                        f"{train_set_name}, feature set {featureset} ...")
        elif task == 'evaluate':
            logger.info(f"Training on {train_set_name}, Test on "
                        f"{test_set_name}, feature set {featureset} ...")
        elif task == 'train':
            logger.info(f"Training on {train_set_name}, feature set "
                        f"{featureset} ...")
        elif task == 'learning_curve':
            logger.info("Generating learning curve "
                        f"({learning_curve_cv_folds} 80/20 folds, "
                        f"sizes={learning_curve_train_sizes}, "
                        f"objective={grid_objective}) on {train_set_name}, "
                        f"feature set {featureset} ...")
        else:  # predict
            logger.info(f"Training on {train_set_name}, Making predictions on"
                        f" {test_set_name}, feature set {featureset} ...")

        # check whether a trained model on the same data with the same
        # featureset already exists if so, load it and then use it on test data
        modelfile = join(model_path, f'{job_name}.model')
        if (task in ['cross_validate', 'learning_curve'] or
                not exists(modelfile) or
                overwrite):
            train_examples = load_featureset(train_path,
                                             featureset,
                                             suffix,
                                             label_col=label_col,
                                             id_col=id_col,
                                             ids_to_floats=ids_to_floats,
                                             quiet=quiet,
                                             class_map=class_map,
                                             feature_hasher=feature_hasher,
                                             num_features=hasher_features,
                                             logger=logger)

            train_set_size = len(train_examples.ids)
            if not train_examples.has_labels:
                raise ValueError('Training examples do not have labels')

            # set up some keyword arguments for instantiating the learner
            # object; note that these are shared by all types of learners
            # supported by SKLL (regular and voting)
            common_learner_kwargs = {"custom_learner_path": custom_learner_path,
                                     "feature_scaling": feature_scaling,
                                     "pos_label": pos_label,
                                     "min_feature_count": min_feature_count,
                                     "logger": logger}

            # instantiate the right type of learner object
            if learner_name in ["VotingClassifier", "VotingRegressor"]:
                # the fixed parameters dictionary must at least
                # contains the estimator names for the voting;
                # the rest can be set to default values
                try:
                    learner_names = fixed_parameters["estimator_names"]
                except KeyError:
                    raise ValueError("'estimator names' must be specified as "
                                     "fixed parameters for voting classifiers "
                                     "and/or regressors.") from None
                else:
                    voting_type = fixed_parameters.get("voting_type", "hard")
                    model_kwargs_list = fixed_parameters.get("estimator_fixed_parameters")
                    sampler_list = fixed_parameters.get("estimator_samplers")
                    sampler_kwargs_list = fixed_parameters.get("estimator_sampler_parameters")
                    param_grids_list = fixed_parameters.get("estimator_param_grids")
                    learner = VotingLearner(learner_names,
                                            voting=voting_type,
                                            model_kwargs_list=model_kwargs_list,
                                            sampler_list=sampler_list,
                                            sampler_kwargs_list=sampler_kwargs_list,
                                            **common_learner_kwargs)
            else:
                learner = Learner(learner_name,
                                  model_kwargs=fixed_parameters,
                                  pipeline=pipeline,
                                  sampler=sampler,
                                  sampler_kwargs=sampler_parameters,
                                  probability=probability,
                                  **common_learner_kwargs)

        # load the model if it already exists
        else:
            # import custom learner into global namespace if we are reusing
            # a saved model
            if custom_learner_path:
                globals()[learner_name] = load_custom_learner(custom_learner_path, learner_name)
            train_set_size = 'unknown'

            # load the non-custom learner from disk
            if exists(modelfile) and not overwrite:
                logger.info(f"Loading pre-existing {learner_name} model: "
                            f"{modelfile}")
            if learner_name in ["VotingClassifier", "VotingRegressor"]:
                learner = VotingLearner.from_file(modelfile, logger=logger)
            else:
                learner = Learner.from_file(modelfile, logger=logger)

        # Load test set if there is one
        if task == 'evaluate' or task == 'predict':
            test_examples = load_featureset(test_path,
                                            featureset,
                                            suffix,
                                            label_col=label_col,
                                            id_col=id_col,
                                            ids_to_floats=ids_to_floats,
                                            quiet=quiet,
                                            class_map=class_map,
                                            feature_hasher=feature_hasher,
                                            num_features=hasher_features)
            test_set_size = len(test_examples.ids)
        else:
            test_set_size = 'n/a'

        # compute information about xval and grid folds that can be put in results
        # in readable form
        if isinstance(cv_folds, dict):
            cv_folds_to_print = f'{len(set(cv_folds.values()))} via folds file'
        else:
            cv_folds_to_print = str(cv_folds)

        if isinstance(grid_search_folds, dict):
            grid_search_folds_to_print = \
                f'{len(set(grid_search_folds.values()))} via folds file'
        else:
            grid_search_folds_to_print = str(grid_search_folds)

        # create a list of dictionaries of the results information
        learner_result_dict_base = {'experiment_name': experiment_name,
                                    'train_set_name': train_set_name,
                                    'train_set_size': train_set_size,
                                    'test_set_name': test_set_name,
                                    'test_set_size': test_set_size,
                                    'featureset': json.dumps(featureset),
                                    'featureset_name': featureset_name,
                                    'shuffle': shuffle,
                                    'learner_name': learner_name,
                                    'task': task,
                                    'start_timestamp':
                                        start_timestamp.strftime('%d %b %Y %H:%M:'
                                                                 '%S.%f'),
                                    'version': __version__,
                                    'feature_scaling': feature_scaling,
                                    'folds_file': folds_file,
                                    'grid_search': grid_search,
                                    'grid_objective': grid_objective,
                                    'grid_search_folds': grid_search_folds_to_print,
                                    'min_feature_count': min_feature_count,
                                    'cv_folds': cv_folds_to_print,
                                    'using_folds_file':
                                        isinstance(cv_folds, dict) or isinstance(grid_search_folds, dict),
                                    'save_cv_folds': save_cv_folds,
                                    'save_cv_models': save_cv_models,
                                    'use_folds_file_for_grid_search': use_folds_file_for_grid_search,
                                    'stratified_folds': stratified_folds,
                                    'scikit_learn_version': SCIKIT_VERSION}

        # check if we're doing cross-validation, because we only load/save
        # models when we're not.
        task_results = None
        if task == 'cross_validate':
            logger.info('Cross-validating')

            # set up the keyword arguments for learner cross-validation;
            # note that most are shared by all types of learners
            # supported by SKLL (regular and voting)
            xval_kwargs = {"shuffle": shuffle,
                           "stratified": stratified_folds,
                           "prediction_prefix": prediction_prefix,
                           "grid_search": grid_search,
                           "grid_search_folds": grid_search_folds,
                           "cv_folds": cv_folds,
                           "cv_seed": cv_seed,
                           "grid_objective": grid_objective,
                           "output_metrics": output_metrics,
                           "grid_jobs": grid_search_jobs,
                           "save_cv_folds": save_cv_folds,
                           "save_cv_models": save_cv_models,
                           "use_custom_folds_for_grid_search": use_folds_file_for_grid_search}

            # voting learners require an optional list of parameter grids
            # passed as fixed parameters whereas regular learners only need
            # a single parameter grid
            if isinstance(learner, VotingLearner):
                xval_kwargs["param_grid_list"] = param_grids_list
                xval_kwargs["individual_predictions"] = save_votes
            else:
                xval_kwargs["param_grid"] = param_grid

            # cross-validate the learner
            results = learner.cross_validate(train_examples, **xval_kwargs)

            # voting learners return only a subset of the results
            # for cross-validation (no grid search results)
            if isinstance(learner, Learner):
                (task_results,
                 grid_scores,
                 grid_search_cv_results_dicts,
                 skll_fold_ids,
                 models) = results
            else:
                (task_results,
                 skll_fold_ids,
                 models) = results
                grid_scores = [None] * cv_folds
                grid_search_cv_results_dicts = [None] * cv_folds

            if models:
                for index, m in enumerate(models, start=1):
                    modelfile = join(model_path, f'{job_name}_fold{index}.model')
                    m.save(modelfile)

        elif task == 'learning_curve':
            logger.info("Generating learning curve(s)")
            (curve_train_scores,
             curve_test_scores,
             computed_curve_train_sizes) = learner.learning_curve(train_examples,
                                                                  grid_objective,
                                                                  cv_folds=learning_curve_cv_folds,
                                                                  train_sizes=learning_curve_train_sizes)
        else:
            # if we do not have a saved model, we need to train one
            grid_scores = [None]
            grid_search_cv_results_dicts = [None]
            if not exists(modelfile) or overwrite:
                logger.info(f"Featurizing and training new {learner_name} "
                            "model")

                # set up the keyword arguments for learner training;
                # note that most are shared by all types of learners
                # supported by SKLL (regular and voting)
                train_kwargs = {"grid_search": grid_search,
                                "grid_search_folds": grid_search_folds,
                                "grid_objective": grid_objective,
                                "grid_jobs": grid_search_jobs,
                                "shuffle": shuffle}

                # voting learners require an optional list of parameter grids
                # passed as fixed parameters whereas regular learners only need
                # a single parameter grid
                if isinstance(learner, VotingLearner):
                    train_kwargs["param_grid_list"] = param_grids_list
                else:
                    train_kwargs["param_grid"] = param_grid

                # train the model
                results = learner.train(train_examples, **train_kwargs)

                # regular learners return a grid score and results
                if isinstance(learner, Learner):
                    grid_scores = [results[0]]
                    grid_search_cv_results_dicts = [results[1]]
                    if grid_search:
                        logger.info(f"Best {grid_objective} grid search score: "
                                    f"{round(results[0], 3)}")

                # save model, if asked
                if model_path:
                    learner.save(modelfile)

            # print out the model parameters; note that for
            # voting learners, we exclude the parameters for
            # the underlying estimators
            params = learner.model.get_params(deep=isinstance(learner, Learner))
            param_out = (f'{param_name}: {param_value}' for
                         param_name, param_value in params.items())
            logger.info(f"Hyperparameters: {', '.join(param_out)}")

            # evaluate the model on the test set;
            if task == 'evaluate':
                logger.info("Evaluating predictions")
                # voting learners have an extra keyword argument indicating
                # whether we want to save the individual predictions (or votes)
                extra_kwargs = {}
                if isinstance(learner, VotingLearner):
                    extra_kwargs["individual_predictions"] = save_votes
                task_results = [learner.evaluate(test_examples,
                                                 prediction_prefix=prediction_prefix,
                                                 grid_objective=grid_objective,
                                                 output_metrics=output_metrics,
                                                 **extra_kwargs)]
            elif task == 'predict':
                logger.info("Writing predictions")
                # we set `class_labels` to `False` so that if the learner is
                # probabilistic, probabilities are written instead of labels;
                # voting learners have an extra keyword argument indicating
                # whether we want to save the individual predictions (or votes)
                extra_kwargs = {}
                if isinstance(learner, VotingLearner):
                    extra_kwargs["individual_predictions"] = save_votes
                learner.predict(test_examples,
                                prediction_prefix=prediction_prefix,
                                class_labels=False,
                                **extra_kwargs)

        end_timestamp = datetime.datetime.now()
        learner_result_dict_base['end_timestamp'] = end_timestamp.strftime(
            '%d %b %Y %H:%M:%S.%f')
        total_time = end_timestamp - start_timestamp
        learner_result_dict_base['total_time'] = str(total_time)

        if task == 'cross_validate' or task == 'evaluate':
            results_json_path = join(results_path,
                                     f'{job_name}.results.json')

            res = _create_learner_result_dicts(task_results,
                                               grid_scores,
                                               grid_search_cv_results_dicts,
                                               learner_result_dict_base)

            # write out the result dictionary to a json file
            with open(results_json_path, 'w') as json_file:
                json.dump(res, json_file, cls=NumpyTypeEncoder)

            with open(join(results_path, f'{job_name}.results'),
                      'w') as output_file:
                _print_fancy_output(res, output_file)

        elif task == 'learning_curve':
            results_json_path = join(results_path, f'{job_name}.results.json')

            res = {}
            res.update(learner_result_dict_base)
            res.update({'learning_curve_cv_folds': learning_curve_cv_folds,
                        'given_curve_train_sizes': learning_curve_train_sizes,
                        'learning_curve_train_scores_means': np.mean(curve_train_scores, axis=1),
                        'learning_curve_test_scores_means': np.mean(curve_test_scores, axis=1),
                        'learning_curve_train_scores_stds': np.std(curve_train_scores, axis=1, ddof=1),
                        'learning_curve_test_scores_stds': np.std(curve_test_scores, axis=1, ddof=1),
                        'computed_curve_train_sizes': computed_curve_train_sizes})

            # we need to return and write out a list of dictionaries
            res = [res]

            # write out the result dictionary to a json file
            with open(results_json_path, 'w') as json_file:
                json.dump(res, json_file, cls=NumpyTypeEncoder)

        # For all other tasks, i.e. train or predict
        else:
            if results_path:
                results_json_path = join(results_path,
                                         f'{job_name}.results.json')

                assert len(grid_scores) == 1
                assert len(grid_search_cv_results_dicts) == 1
                grid_search_cv_results_dict = {"grid_score": grid_scores[0]}
                grid_search_cv_results_dict["grid_search_cv_results"] = \
                    grid_search_cv_results_dicts[0]
                grid_search_cv_results_dict.update(learner_result_dict_base)
                # write out the result dictionary to a json file
                with open(results_json_path, 'w') as json_file:
                    json.dump(grid_search_cv_results_dict, json_file, cls=NumpyTypeEncoder)
            res = [learner_result_dict_base]

        # write out the cv folds if required
        if task == 'cross_validate' and save_cv_folds:
            skll_fold_ids_file = f'{experiment_name}_skll_fold_ids.csv'
            with open(join(results_path, skll_fold_ids_file),
                      'w') as output_file:
                _write_skll_folds(skll_fold_ids, output_file)

    finally:
        close_and_remove_logger_handlers(logger)

    return res


def run_configuration(config_file, local=False, overwrite=True, queue='all.q',  # noqa: C901
                      hosts=None, write_summary=True, quiet=False,
                      ablation=0, resume=False, log_level=logging.INFO):
    """
    Takes a configuration file and runs the specified jobs on the grid.

    Parameters
    ----------
    config_file : str
        Path to the configuration file we would like to use.
    local : bool, optional
        Should this be run locally instead of on the cluster?
        Defaults to ``False``.
    overwrite : bool, optional
        If the model files already exist, should we overwrite
        them instead of re-using them?
        Defaults to ``True``.
    queue : str, optional
        The DRMAA queue to use if we're running on the cluster.
        Defaults to ``'all.q'``.
    hosts : list of str, optional
        If running on the cluster, these are the machines we should use.
        Defaults to ``None``.
    write_summary : bool, optional
        Write a TSV file with a summary of the results.
        Defaults to ``True``.
    quiet : bool, optional
        Suppress printing of "Loading..." messages.
        Defaults to ``False``.
    ablation : int, optional
        Number of features to remove when doing an ablation
        experiment. If positive, we will perform repeated ablation
        runs for all combinations of features removing the
        specified number at a time. If ``None``, we will use all
        combinations of all lengths. If 0, the default, no
        ablation is performed. If negative, a ``ValueError`` is
        raised.
        Defaults to 0.
    resume : bool, optional
        If result files already exist for an experiment, do not
        overwrite them. This is very useful when doing a large
        ablation experiment and part of it crashes.
        Defaults to ``False``.
    log_level : str, optional
        The level for logging messages.
        Defaults to ``logging.INFO``.

    Returns
    -------
    result_json_paths : list of str
        A list of paths to .json results files for each variation in the
        experiment.

    Raises
    ------
    ValueError
        If value for ``"ablation"`` is not a positive int or ``None``.
    OSError
        If the lenth of the ``FeatureSet`` name > 210.
    """

    try:

        # Read configuration
        (experiment_name, task, sampler, fixed_sampler_parameters, feature_hasher,
         hasher_features, id_col, label_col, train_set_name, test_set_name, suffix,
         featuresets, do_shuffle, model_path, do_grid_search, grid_objectives,
         probability, pipeline, results_path, pos_label, feature_scaling,
         min_feature_count, folds_file, grid_search_jobs, grid_search_folds, cv_folds,
         cv_seed, save_cv_folds, save_cv_models, use_folds_file_for_grid_search,
         do_stratified_folds, fixed_parameter_list, param_grid_list, featureset_names,
         learners, prediction_dir, log_path, train_path, test_path, ids_to_floats,
         class_map, custom_learner_path, custom_metric_path,
         learning_curve_cv_folds_list, learning_curve_train_sizes,
         output_metrics, save_votes) = parse_config_file(config_file, log_level=log_level)

        # get the main experiment logger that will already have been
        # created by the configuration parser so we don't need anything
        # except the name `experiment`.
        logger = get_skll_logger('experiment')

        # Check if we have gridmap
        if not local and not _HAVE_GRIDMAP:
            local = True
            logger.warning('gridmap 0.10.1+ not available. Forcing local '
                           'mode.  To run things on a DRMAA-compatible '
                           'cluster, install gridmap>=0.10.1 via pip.')

        # No grid search or ablation for learning curve generation
        if task == 'learning_curve':
            if ablation is None or ablation > 0:
                ablation = 0
                logger.warning("Ablating features is not supported during "
                               "learning curve generation. Ignoring.")

        # if we just had a train file and a test file, there are no real featuresets
        # in which case there are no features to ablate
        if len(featuresets) == 1 and len(featuresets[0]) == 1:
            if ablation is None or ablation > 0:
                ablation = 0
                logger.warning("Not enough featuresets for ablation. Ignoring.")

        # if performing ablation, expand featuresets to include combinations of
        # features within those sets
        if ablation is None or ablation > 0:
            # Make new feature set lists so that we can iterate without issue
            expanded_fs = []
            expanded_fs_names = []
            for features, featureset_name in zip(featuresets, featureset_names):
                features = sorted(features)
                featureset = set(features)
                # Expand to all feature combinations if ablation is None
                if ablation is None:
                    for i in range(1, len(features)):
                        for excluded_features in combinations(features, i):
                            expanded_fs.append(sorted(featureset -
                                                      set(excluded_features)))
                            expanded_fs_names.append(
                                f'{featureset_name}_minus_'
                                f'{_munge_featureset_name(excluded_features)}'
                            )
                # Otherwise, just expand removing the specified number at a time
                else:
                    for excluded_features in combinations(features, ablation):
                        expanded_fs.append(sorted(featureset -
                                                  set(excluded_features)))
                        expanded_fs_names.append(
                            f'{featureset_name}_minus_'
                            f'{_munge_featureset_name(excluded_features)}'
                        )
                # Also add version with nothing removed as baseline
                expanded_fs.append(features)
                expanded_fs_names.append(f'{featureset_name}_all')

            # Replace original feature set lists
            featuresets = expanded_fs
            featureset_names = expanded_fs_names
        elif ablation < 0:
            raise ValueError('Value for "ablation" argument must be either '
                             'positive integer or None.')

        # the list of jobs submitted (if running on grid)
        if not local:
            jobs = []

        # the list to hold the paths to all the result json files
        result_json_paths = []

        # check if the length of the featureset_name exceeds the maximum length
        # allowed
        for featureset_name in featureset_names:
            if len(featureset_name) > 210:
                raise OSError(
                    f'System generated file length "{featureset_name}" '
                    'exceeds the maximum length supported.  Please specify '
                    'names of your datasets with "featureset_names".  If you '
                    'are running ablation experiment, please reduce the '
                    'length of the features in "featuresets" because the '
                    'auto-generated name would be longer than the file system'
                    ' can handle')

        # if the task is learning curve, and ``metrics`` was specified, then
        # assign the value of ``metrics`` to ``grid_objectives`` - this lets
        # us piggyback on the parallelization of the objectives that is already
        # set up for us to use
        if task == 'learning_curve' and len(output_metrics) > 0:
            grid_objectives = output_metrics

        # if there were no grid objectives provided, just set it to
        # a list containing a single None so as to allow the parallelization
        # to proceeed and to pass the correct default value of grid_objective
        # down to _classify_featureset().
        if not grid_objectives:
            grid_objectives = [None]

        # Run each featureset-learner-objective combination
        for featureset, featureset_name in zip(featuresets, featureset_names):
            for learner_num, learner_name in enumerate(learners):
                for grid_objective in grid_objectives:

                    # for the individual job name, we need to add the feature set name
                    # and the learner name
                    if grid_objective is None or len(grid_objectives) == 1:
                        job_name_components = [experiment_name, featureset_name,
                                               learner_name]
                    else:
                        job_name_components = [experiment_name, featureset_name,
                                               learner_name, grid_objective]

                    job_name = '_'.join(job_name_components)

                    # change the prediction prefix to include the feature set
                    prediction_prefix = join(prediction_dir, job_name)

                    # the log file that stores the actual output of this script (e.g.,
                    # the tuned parameters, what kind of experiment was run, etc.)
                    logfile = join(log_path, f'{job_name}.log')

                    # Figure out result json file path
                    result_json_path = join(results_path,
                                            f'{job_name}.results.json')

                    # save the path to the results json file that will be written
                    result_json_paths.append(result_json_path)

                    # If result file already exists and we're resuming, move on
                    if resume and (exists(result_json_path) and getsize(result_json_path)):
                        logger.info('Running in resume mode and '
                                    f'{result_json_path} exists, so skipping '
                                    'job.')
                        continue

                    # create job if we're doing things on the grid
                    job_args = {}
                    job_args["experiment_name"] = experiment_name
                    job_args["task"] = task
                    job_args["sampler"] = sampler
                    job_args["feature_hasher"] = feature_hasher
                    job_args["hasher_features"] = hasher_features
                    job_args["job_name"] = job_name
                    job_args["featureset"] = featureset
                    job_args["featureset_name"] = featureset_name
                    job_args["learner_name"] = learner_name
                    job_args["train_path"] = train_path
                    job_args["test_path"] = test_path
                    job_args["train_set_name"] = train_set_name
                    job_args["test_set_name"] = test_set_name
                    job_args["shuffle"] = do_shuffle
                    job_args["model_path"] = model_path
                    job_args["prediction_prefix"] = prediction_prefix
                    job_args["grid_search"] = do_grid_search
                    job_args["grid_objective"] = grid_objective
                    job_args['output_metrics'] = output_metrics
                    job_args["suffix"] = suffix
                    job_args["log_file"] = logfile
                    job_args["log_level"] = log_level
                    job_args["probability"] = probability
                    job_args["pipeline"] = pipeline
                    job_args["save_votes"] = save_votes
                    job_args["results_path"] = results_path
                    job_args["sampler_parameters"] = (fixed_sampler_parameters
                                                      if fixed_sampler_parameters
                                                      else dict())
                    job_args["fixed_parameters"] = (fixed_parameter_list[learner_num]
                                                    if fixed_parameter_list
                                                    else dict())
                    job_args["param_grid"] = (param_grid_list[learner_num]
                                              if param_grid_list else None)
                    job_args["pos_label"] = pos_label
                    job_args["overwrite"] = overwrite
                    job_args["feature_scaling"] = feature_scaling
                    job_args["min_feature_count"] = min_feature_count
                    job_args["grid_search_jobs"] = grid_search_jobs
                    job_args["grid_search_folds"] = grid_search_folds
                    job_args["folds_file"] = folds_file
                    job_args["cv_folds"] = cv_folds
                    job_args["cv_seed"] = cv_seed
                    job_args["save_cv_folds"] = save_cv_folds
                    job_args["save_cv_models"] = save_cv_models
                    job_args["use_folds_file_for_grid_search"] = use_folds_file_for_grid_search
                    job_args["do_stratified_folds"] = do_stratified_folds
                    job_args["label_col"] = label_col
                    job_args["id_col"] = id_col
                    job_args["ids_to_floats"] = ids_to_floats
                    job_args["quiet"] = quiet
                    job_args["class_map"] = class_map
                    job_args["custom_learner_path"] = custom_learner_path
                    job_args["custom_metric_path"] = custom_metric_path
                    job_args["learning_curve_cv_folds"] = learning_curve_cv_folds_list[learner_num]
                    job_args["learning_curve_train_sizes"] = learning_curve_train_sizes

                    if not local:
                        jobs.append(Job(_classify_featureset,
                                        [job_args],
                                        num_slots=(MAX_CONCURRENT_PROCESSES if
                                                   (do_grid_search or task == 'learning_curve') else 1),
                                        name=job_name,
                                        queue=queue))
                    else:
                        _classify_featureset(job_args)

        # Call get_skll_logger again after _classify_featureset
        # calls are finished so that any warnings that may
        # happen after this point get correctly logged to the
        # main logger
        logger = get_skll_logger('experiment')

        # submit the jobs (if running on grid)
        if not local and _HAVE_GRIDMAP:
            if log_path:
                job_results = process_jobs(jobs, white_list=hosts,
                                           temp_dir=log_path)
            else:
                job_results = process_jobs(jobs, white_list=hosts)
            _check_job_results(job_results)

        # write out the summary results file
        if (task == 'cross_validate' or task == 'evaluate') and write_summary:
            summary_file_name = f'{experiment_name}_summary.tsv'
            with open(join(results_path,
                           summary_file_name), 'w', newline='') as output_file:
                _write_summary_file(result_json_paths,
                                    output_file,
                                    ablation=ablation)
        elif task == 'learning_curve':
            output_file_name = f'{experiment_name}_summary.tsv'
            output_file_path = join(results_path, output_file_name)
            with open(output_file_path, 'w', newline='') as output_file:
                _write_learning_curve_file(result_json_paths, output_file)

            # generate the actual plot if we have the requirements installed
            generate_learning_curve_plots(experiment_name,
                                          results_path,
                                          output_file_path)

    finally:

        # Close/remove any logger handlers
        close_and_remove_logger_handlers(get_skll_logger('experiment'))

    return result_json_paths



# License: BSD 3 clause
"""
Utility classes and functions to parse SKLL configuration files.

:author: Nitin Madnani (nmadnani@ets.org)
:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
"""

import csv
import errno
import logging
from os.path import exists, isabs, join, normpath

import ruamel.yaml as yaml


def fix_json(json_string):
    """
    Fixes incorrectly formatted quotes and capitalized booleans in the given
    JSON string.

    Parameters
    ----------
    json_string : str
        A JSON-style string.

    Returns
    -------
    json_string : str
        The normalized JSON string.
    """
    json_string = json_string.replace('True', 'true')
    json_string = json_string.replace('False', 'false')
    json_string = json_string.replace("'", '"')
    return json_string


def load_cv_folds(folds_file, ids_to_floats=False):
    """
    Loads cross-validation folds from a CSV file with two columns for example
    ID and fold ID (and a header).

    Parameters
    ----------
    folds_file : str
        The path to a folds file to read.

    ids_to_floats : bool, default=False
        Whether to convert IDs to floats.

    Returns
    -------
    res : dict
        A dictionary with example IDs as the keys and fold IDs as the values.

    Raises
    ------
    ValueError
        If example IDs cannot be converted to floats and `ids_to_floats` is `True`.
    """
    with open(folds_file) as f:
        reader = csv.reader(f)
        next(reader)  # discard the header
        res = {}
        for row in reader:
            if ids_to_floats:
                try:
                    row[0] = float(row[0])
                except ValueError:
                    raise ValueError('You set ids_to_floats to true, but ID '
                                     f'{row[0]} could not be converted to '
                                     'float')
            res[row[0]] = row[1]

    return res


def locate_file(file_path, config_dir):
    """
    Locate a file, given a file path and configuration directory.

    Parameters
    ----------
    file_path : str
        The file to locate. Path may be absolute or relative.

    config_dir : str
        The path to the configuration file directory.

    Returns
    -------
    path_to_check : str
        The normalized absolute path, if it exists.

    Raises
    ------
    IOError
        If the file does not exist.
    """
    if not file_path:
        return ''
    path_to_check = file_path if isabs(file_path) else normpath(join(config_dir,
                                                                     file_path))
    ans = exists(path_to_check)
    if not ans:
        raise IOError(errno.ENOENT, "File does not exist", path_to_check)
    else:
        return path_to_check


def _munge_featureset_name(featureset):
    """
    Joins features in ``featureset`` with a '+' if ``featureset`` is not a string.
    Otherwise, returns ``featureset``.

    Parameters
    ----------
    featureset : skll.data.FeatureSet
        A SKLL ``FeatureSet`` object.

    Returns
    -------
    res : str
        ``featureset`` names joined with '+', if ``featureset`` is not a string.
    """
    if isinstance(featureset, str):
        return featureset

    res = '+'.join(sorted(featureset))
    return res


def _parse_and_validate_metrics(metrics, option_name, logger=None):
    """
    Given a string containing a list of metrics, this function
    parses that string into a list and validates some specific
    metric names.

    Parameters
    ----------
    metrics : str
        A string containing a list of metrics.

    option_name : str
        The name of the option with which the metrics are associated.

    logger : logging.Logger, default=None
        A logging object.

    Returns
    -------
    metrics : list of str
        A list of metrics for the given option.

    Raises
    ------
    TypeError
        If the given string cannot be converted to a list.

    ValueError
        If "mean_squared_error" is specified as a metric.
    """

    # create a logger if one was not passed in
    if not logger:
        logger = logging.getLogger(__name__)

    # make sure the given metrics data type is a list
    # and parse it correctly
    metrics = yaml.safe_load(fix_json(metrics))
    if not isinstance(metrics, list):
        raise TypeError(f"{option_name} should be a list, not a "
                        f"{type(metrics)}.")

    # `mean_squared_error` is no longer supported.
    # It has been replaced by `neg_mean_squared_error`
    if 'mean_squared_error' in metrics:
        raise ValueError("The metric \"mean_squared_error\" is no longer "
                         "supported. please use the metric "
                         "\"neg_mean_squared_error\" instead.")

    return metrics



# License: BSD 3 clause
"""
Functions related to running experiments and parsing configuration files.

:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:author: Nitin Madnani (nmadnani@ets.org)
:author: Chee Wee Leong (cleong@ets.org)
"""

import csv
import json
import math
import sys
from collections import defaultdict
from os.path import exists, join

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import ruamel.yaml as yaml
import seaborn as sns

from skll.utils.logging import get_skll_logger

# Turn off interactive plotting for matplotlib
plt.ioff()


def _compute_ylimits_for_featureset(df, metrics):
    """
    Compute the y-limits for learning curve plots.

    Parameters
    ----------
    df : pd.DataFrame
        A data_frame with relevant metric information for
        train and test.
    metrics : list of str
        A list of metrics for learning curve plots.

    Returns
    -------
    ylimits : dict
        A dictionary, with metric names as keys
        and a tuple of (lower_limit, upper_limit) as values.
    """

    # set the y-limits of the curves depending on what kind
    # of values the metric produces
    ylimits = {}
    for metric in metrics:
        # get the real min and max for the values that will be plotted
        df_train = df[(df['variable'] == 'train_score_mean') & (df['metric'] == metric)]
        df_test = df[(df['variable'] == 'test_score_mean') & (df['metric'] == metric)]
        train_values_lower = df_train['value'].values - df_train['train_score_std'].values
        test_values_lower = df_test['value'].values - df_test['test_score_std'].values
        min_score = np.min(np.concatenate([train_values_lower,
                                           test_values_lower]))
        train_values_upper = df_train['value'].values + df_train['train_score_std'].values
        test_values_upper = df_test['value'].values + df_test['test_score_std'].values
        max_score = np.max(np.concatenate([train_values_upper,
                                           test_values_upper]))

        # squeeze the limits to hide unnecessary parts of the graph
        # set the limits with a little buffer on either side but not too much
        if min_score < 0:
            lower_limit = max(min_score - 0.1, math.floor(min_score) - 0.05)
        else:
            lower_limit = 0

        if max_score > 0:
            upper_limit = min(max_score + 0.1, math.ceil(max_score) + 0.05)
        else:
            upper_limit = 0

        ylimits[metric] = (lower_limit, upper_limit)

    return ylimits


def generate_learning_curve_plots(experiment_name,
                                  output_dir,
                                  learning_curve_tsv_file):
    """
    Generate the learning curve plots given the TSV output
    file from a learning curve experiment.

    Parameters
    ----------
    experiment_name : str
        The name of the experiment.
    output_dir : str
        Path to the output directory for the plots.
    learning_curve_tsv_file : str
        The path to the learning curve TSV file.
    """

    # use pandas to read in the TSV file into a data frame
    # and massage it from wide to long format for plotting
    df = pd.read_csv(learning_curve_tsv_file, sep='\t')
    num_learners = len(df['learner_name'].unique())
    num_metrics = len(df['metric'].unique())
    df_melted = pd.melt(df, id_vars=[c for c in df.columns
                                     if c not in ['train_score_mean', 'test_score_mean']])

    # if there are any training sizes greater than 1000,
    # then we should probably rotate the tick labels
    # since otherwise the labels are not clearly rendered
    rotate_labels = np.any([size >= 1000 for size in df['training_set_size'].unique()])

    # set up and draw the actual learning curve figures, one for
    # each of the featuresets
    for fs_name, df_fs in df_melted.groupby('featureset_name'):
        fig = plt.figure()
        fig.set_size_inches(2.5 * num_learners, 2.5 * num_metrics)

        # compute ylimits for this feature set for each objective
        with sns.axes_style('whitegrid', {"grid.linestyle": ':',
                                          "xtick.major.size": 3.0}):
            g = sns.FacetGrid(df_fs, row="metric", col="learner_name",
                              hue="variable", height=2.5, aspect=1,
                              margin_titles=True, despine=True, sharex=False,
                              sharey=False, legend_out=False, palette="Set1")
            colors = train_color, test_color = sns.color_palette("Set1")[:2]
            g = g.map_dataframe(sns.pointplot, "training_set_size", "value",
                                scale=.5, ci=None)
            ylimits = _compute_ylimits_for_featureset(df_fs, g.row_names)
            for ax in g.axes.flat:
                plt.setp(ax.texts, text="")
            g = (g.set_titles(row_template='', col_template='{col_name}')
                 .set_axis_labels('Training Examples', 'Score'))
            if rotate_labels:
                g = g.set_xticklabels(rotation=60)

            for i, row_name in enumerate(g.row_names):
                for j, col_name in enumerate(g.col_names):
                    ax = g.axes[i][j]
                    ax.set(ylim=ylimits[row_name])
                    df_ax_train = df_fs[(df_fs['learner_name'] == col_name) &
                                        (df_fs['metric'] == row_name) &
                                        (df_fs['variable'] == 'train_score_mean')]
                    df_ax_test = df_fs[(df_fs['learner_name'] == col_name) &
                                       (df_fs['metric'] == row_name) &
                                       (df_fs['variable'] == 'test_score_mean')]
                    ax.fill_between(list(range(len(df_ax_train))),
                                    df_ax_train['value'] - df_ax_train['train_score_std'],
                                    df_ax_train['value'] + df_ax_train['train_score_std'],
                                    alpha=0.1,
                                    color=train_color)
                    ax.fill_between(list(range(len(df_ax_test))),
                                    df_ax_test['value'] - df_ax_test['test_score_std'],
                                    df_ax_test['value'] + df_ax_test['test_score_std'],
                                    alpha=0.1,
                                    color=test_color)
                    if j == 0:
                        ax.set_ylabel(row_name)
                        if i == 0:
                            # set up the legend handles for this plot
                            plot_handles = [matplotlib.lines.Line2D([],
                                                                    [],
                                                                    color=c,
                                                                    label=l,
                                                                    linestyle='-')
                                            for c, l in zip(colors, ['Training',
                                                                     'Cross-validation'])]
                            ax.legend(handles=plot_handles,
                                      loc=4,
                                      fancybox=True,
                                      fontsize='x-small',
                                      ncol=1,
                                      frameon=True)
            g.fig.tight_layout(w_pad=1)
            plt.savefig(join(output_dir, f'{experiment_name}_{fs_name}.png'),
                        dpi=300)
            # explicitly close figure to save memory
            plt.close(fig)


def _print_fancy_output(learner_result_dicts, output_file=sys.stdout):
    """
    Function to take all of the results from all of the folds and print
    nice tables with the results.

    Parameters
    ----------
    learner_result_dicts : list of str
        A list of paths to the individual result JSON files.
    output_file : file buffer, optional
        The file buffer to print to.
        Defaults to ``sys.stdout``.
    """
    if not learner_result_dicts:
        raise ValueError('Result dictionary list is empty!')

    lrd = learner_result_dicts[0]
    print(f'Experiment Name: {lrd["experiment_name"]}', file=output_file)
    print(f'SKLL Version: {lrd["version"]}', file=output_file)
    print(f'Training Set: {lrd["train_set_name"]}', file=output_file)
    print(f'Training Set Size: {lrd["train_set_size"]}', file=output_file)
    print(f'Test Set: {lrd["test_set_name"]}', file=output_file)
    print(f'Test Set Size: {lrd["test_set_size"]}', file=output_file)
    print(f'Shuffle: {lrd["shuffle"]}', file=output_file)
    print(f'Feature Set: {lrd["featureset"]}', file=output_file)
    print(f'Learner: {lrd["learner_name"]}', file=output_file)
    print(f'Task: {lrd["task"]}', file=output_file)
    if lrd['folds_file']:
        print(f'Specified Folds File: {lrd["folds_file"]}', file=output_file)
    if lrd['task'] == 'cross_validate':
        print(f'Number of Folds: {lrd["cv_folds"]}', file=output_file)
        if not lrd['cv_folds'].endswith('folds file'):
            print(f'Stratified Folds: {lrd["stratified_folds"]}',
                  file=output_file)
    print(f'Feature Scaling: {lrd["feature_scaling"]}', file=output_file)
    print(f'Grid Search: {lrd["grid_search"]}', file=output_file)
    if lrd['grid_search']:
        print(f'Grid Search Folds: {lrd["grid_search_folds"]}',
              file=output_file)
        print(f'Grid Objective Function: {lrd["grid_objective"]}',
              file=output_file)
    if (lrd['task'] == 'cross_validate' and
            lrd['grid_search'] and
            lrd['cv_folds'].endswith('folds file')):
        print('Using Folds File for Grid Search: '
              f'{lrd["use_folds_file_for_grid_search"]}',
              file=output_file)
    if lrd['task'] in ['evaluate', 'cross_validate'] and lrd['additional_scores']:
        print('Additional Evaluation Metrics: '
              f'{list(lrd["additional_scores"].keys())}',
              file=output_file)
    print(f'Scikit-learn Version: {lrd["scikit_learn_version"]}',
          file=output_file)
    print(f'Start Timestamp: {lrd["start_timestamp"]}', file=output_file)
    print(f'End Timestamp: {lrd["end_timestamp"]}', file=output_file)
    print(f'Total Time: {lrd["total_time"]}', file=output_file)
    print('\n', file=output_file)

    for lrd in learner_result_dicts:
        print(f'Fold: {lrd["fold"]}', file=output_file)
        print(f'Model Parameters: {lrd.get("model_params", "")}',
              file=output_file)
        print(f'Grid Objective Score (Train) = {lrd.get("grid_score", "")}',
              file=output_file)
        if 'result_table' in lrd:
            print(lrd['result_table'], file=output_file)
            print(f'Accuracy = {lrd["accuracy"]}', file=output_file)
        if 'descriptive' in lrd:
            print('Descriptive statistics:', file=output_file)
            for desc_stat in ['min', 'max', 'avg', 'std']:
                actual = lrd['descriptive']['actual'][desc_stat]
                predicted = lrd['descriptive']['predicted'][desc_stat]
                print(f' {desc_stat.title()} = {actual:.4f} (actual), '
                      f'{predicted:.4f} (predicted)',
                      file=output_file)
            print(f'Pearson = {lrd["pearson"]:f}', file=output_file)
        print(f'Objective Function Score (Test) = {lrd.get("score", "")}',
              file=output_file)

        # now print the additional metrics, if there were any
        if lrd['additional_scores']:
            print('', file=output_file)
            print('Additional Evaluation Metrics (Test):', file=output_file)
            for metric, score in lrd['additional_scores'].items():
                score = '' if np.isnan(score) else score
                print(f' {metric} = {score}', file=output_file)
        print('', file=output_file)


def _write_learning_curve_file(result_json_paths, output_file):
    """
    Function to take a list of paths to individual learning curve
    results json files and writes out a single TSV file with the
    learning curve data.

    Parameters
    ----------
    result_json_paths : list of str
        A list of paths to the individual result JSON files.
    output_file : str
        The path to the output file (TSV format).
    """

    learner_result_dicts = []

    # Map from feature set names to all features in them
    logger = get_skll_logger('experiment')
    for json_path in result_json_paths:
        if not exists(json_path):
            logger.error(f'JSON results file {json_path} not found. Skipping '
                         'summary creation. You can manually create the '
                         'summary file after the fact by using the '
                         'summarize_results script.')
            return
        else:
            with open(json_path) as json_file:
                obj = json.load(json_file)
                learner_result_dicts.extend(obj)

    # Build and write header
    header = ['featureset_name', 'learner_name', 'metric',
              'train_set_name', 'training_set_size', 'train_score_mean',
              'test_score_mean', 'train_score_std', 'test_score_std',
              'scikit_learn_version', 'version']
    writer = csv.DictWriter(output_file,
                            header,
                            extrasaction='ignore',
                            dialect=csv.excel_tab)
    writer.writeheader()

    # write out the fields we need for the learning curve file
    # specifically, we need to separate out the curve sizes
    # and scores into individual entries.
    for lrd in learner_result_dicts:
        training_set_sizes = lrd['computed_curve_train_sizes']
        train_scores_means_by_size = lrd['learning_curve_train_scores_means']
        test_scores_means_by_size = lrd['learning_curve_test_scores_means']
        train_scores_stds_by_size = lrd['learning_curve_train_scores_stds']
        test_scores_stds_by_size = lrd['learning_curve_test_scores_stds']

        # rename `grid_objective` to `metric` since the latter name can be confusing
        lrd['metric'] = lrd['grid_objective']

        for (size,
             train_score_mean,
             test_score_mean,
             train_score_std,
             test_score_std) in zip(training_set_sizes,
                                    train_scores_means_by_size,
                                    test_scores_means_by_size,
                                    train_scores_stds_by_size,
                                    test_scores_stds_by_size):
            lrd['training_set_size'] = size
            lrd['train_score_mean'] = train_score_mean
            lrd['test_score_mean'] = test_score_mean
            lrd['train_score_std'] = train_score_std
            lrd['test_score_std'] = test_score_std

            writer.writerow(lrd)

    output_file.flush()


def _write_skll_folds(skll_fold_ids, skll_fold_ids_file):
    """
    Function to take a dictionary of id->test-fold-number and
    write it to a file.

    Parameters
    ----------
    skll_fold_ids : dict
        Dictionary with ids as keys and test-fold-numbers as values.
    skll_fold_ids_file : file buffer
        An open file handler to write to.
    """

    f = csv.writer(skll_fold_ids_file)
    f.writerow(['id', 'cv_test_fold'])
    for example_id in skll_fold_ids:
        f.writerow([example_id, skll_fold_ids[example_id]])

    skll_fold_ids_file.flush()


def _write_summary_file(result_json_paths, output_file, ablation=0):
    """
    Function to take a list of paths to individual result
    json files and returns a single file that summarizes
    all of them.

    Parameters
    ----------
    result_json_paths : list of str
        A list of paths to the individual result JSON files.
    output_file : str
        The path to the output file (TSV format).
    ablation : int, optional
        The number of features to remove when doing ablation experiment.
        Defaults to 0.
    """
    learner_result_dicts = []
    # Map from feature set names to all features in them
    all_features = defaultdict(set)
    logger = get_skll_logger('experiment')
    for json_path in result_json_paths:
        if not exists(json_path):
            logger.error(f'JSON results file {json_path} not found. Skipping '
                         'summary creation. You can manually create the '
                         'summary file after the fact by using the '
                         'summarize_results script.')
            return
        else:
            with open(json_path) as json_file:
                obj = json.load(json_file)
                featureset_name = obj[0]['featureset_name']
                if ablation != 0 and '_minus_' in featureset_name:
                    parent_set = featureset_name.split('_minus_', 1)[0]
                    all_features[parent_set].update(
                        yaml.safe_load(obj[0]['featureset']))
                learner_result_dicts.extend(obj)

    # Build and write header
    header = set(learner_result_dicts[0].keys()) - {'result_table',
                                                    'descriptive'}
    if ablation != 0:
        header.add('ablated_features')
    header = sorted(header)
    writer = csv.DictWriter(output_file,
                            header,
                            extrasaction='ignore',
                            dialect=csv.excel_tab)
    writer.writeheader()

    # Build "ablated_features" list and fix some backward compatible things
    for lrd in learner_result_dicts:
        featureset_name = lrd['featureset_name']
        if ablation != 0:
            parent_set = featureset_name.split('_minus_', 1)[0]
            ablated_features = all_features[parent_set].difference(
                yaml.safe_load(lrd['featureset']))
            lrd['ablated_features'] = ''
            if ablated_features:
                lrd['ablated_features'] = json.dumps(sorted(ablated_features))

        # write out the new learner dict with the readable fields
        writer.writerow(lrd)

    output_file.flush()



#!/usr/bin/env python
# License: BSD 3 clause
'''
Filter a given feature file to remove unwanted features, labels, or IDs.

:author: Dan Blanchard (dblanchard@ets.org)
:author: Nitin Madnani (nmadnani@ets.org)
:organization: ETS
'''

import argparse
import logging
import os
import sys

from skll.data.readers import EXT_TO_READER, safe_float
from skll.data.writers import EXT_TO_WRITER, ARFFWriter, CSVWriter, TSVWriter
from skll.version import __version__


def main(argv=None):
    """
    Handles command line arguments and gets things started.

    Parameters
    ----------
    argv : list of str
        List of arguments, as if specified on the command-line.
        If None, ``sys.argv[1:]`` is used instead.
    """

    # Get command line arguments
    parser = argparse.ArgumentParser(
        description="Takes an input feature file and removes any instances or"
                    " features that do not match the specified patterns.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument('-i', '--input',
                        dest='infile',
                        required=True,
                        help='input feature file (ends in .arff, .csv, '
                             '.jsonlines, .ndj, or .tsv)')
    parser.add_argument('-o', '--output',
                        dest='outfile',
                        required=True,
                        help='output feature file (must have same extension '
                             'as input file)')
    parser.add_argument('-f', '--feature',
                        help='A feature in the feature file you would like to '
                             'keep.  If unspecified, no features are removed.',
                        nargs='*')
    parser.add_argument('-I', '--id',
                        help='An instance ID in the feature file you would '
                             'like to keep.  If unspecified, no instances are'
                             ' removed based on their IDs.',
                        nargs='*')
    parser.add_argument('--id_col',
                        help='Name of the column which contains the instance '
                             'IDs in ARFF, CSV, or TSV files.',
                        default='id')
    parser.add_argument('--inverse',
                        help='Instead of keeping features and/or examples in '
                             'lists, remove them.',
                        action='store_true')
    parser.add_argument('-L', '--label',
                        help='A label in the feature file you would like to '
                             'keep.  If unspecified, no instances are removed'
                             ' based on their labels.',
                        nargs='*')
    parser.add_argument('-l', '--label_col',
                        help='Name of the column which contains the class '
                             'labels in ARFF, CSV, or TSV files. For ARFF '
                             'files, this must be the final column to count '
                             'as the label.',
                        default='y')
    parser.add_argument('-rb', '--replace_blanks_with',
                        help='Specifies a new value with which to replace '
                             'blank values in all columns in a CSV/TSV file. To '
                             'replace blanks differently in each column, use '
                             'the SKLL Reader API directly.',
                        default=None)
    parser.add_argument('-db', '--drop_blanks',
                        action='store_true',
                        help='Drop all lines/rows in a CSV/TSV file that '
                             'contain any blank values. If each row of the file '
                             'contains a blank value, an exception will be raised.',
                        default=False)
    parser.add_argument('-q', '--quiet',
                        help='Suppress printing of "Loading..." messages.',
                        action='store_true')
    parser.add_argument('--version', action='version',
                        version=f'%(prog)s {__version__}')
    args = parser.parse_args(argv)

    # Make warnings from built-in warnings module get formatted more nicely
    logging.captureWarnings(True)
    logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - '
                               '%(message)s')
    logger = logging.getLogger(__name__)

    # all extensions except .libsvm can be processed
    valid_extensions = {ext for ext in EXT_TO_READER if ext != '.libsvm'}

    # make sure the input file extension is one we can process
    input_extension = os.path.splitext(args.infile)[1].lower()
    output_extension = os.path.splitext(args.outfile)[1].lower()

    if input_extension == '.libsvm':
        logger.error('Cannot filter LibSVM files.  Please use skll_convert to'
                     ' convert to a different datatype first.')
        sys.exit(1)

    if input_extension not in valid_extensions:
        logger.error('Input file must be in either .arff, .csv, .jsonlines, '
                     '.ndj, or .tsv format. You specified: '
                     f'{input_extension}')
        sys.exit(1)

    if output_extension != input_extension:
        logger.error('Output file must be in the same format as the input '
                     f'file.  You specified: {output_extension}')
        sys.exit(1)

    if input_extension == '.csv' or input_extension == '.tsv':
        replace_blanks_with = args.replace_blanks_with
        drop_blanks = args.drop_blanks
        if drop_blanks and replace_blanks_with is not None:
            raise ValueError("You cannot both drop blanks and replace them. "
                             "'replace_blanks_with' can only have a value "
                             "when 'drop_blanks' is `False`.")
        replace_blanks_with = (None if replace_blanks_with is None
                               else safe_float(replace_blanks_with))
        kwargs = {'replace_blanks_with': replace_blanks_with,
                  'drop_blanks': drop_blanks}
    else:
        kwargs = {}

    # Read input file
    reader = EXT_TO_READER[input_extension](args.infile,
                                            quiet=args.quiet,
                                            label_col=args.label_col,
                                            id_col=args.id_col,
                                            **kwargs)
    feature_set = reader.read()

    # convert the specified labels with `safe_float` before using
    # to match what the various featureset readers do
    converted_labels = None
    if args.label:
        converted_labels = [safe_float(label) for label in args.label]

    # Do the actual filtering
    feature_set.filter(ids=args.id,
                       labels=converted_labels,
                       features=args.feature,
                       inverse=args.inverse)

    # write out the file in the requested output format
    writer_type = EXT_TO_WRITER[input_extension]
    writer_args = {'quiet': args.quiet}
    if writer_type is CSVWriter or writer_type is TSVWriter:
        writer_args['label_col'] = args.label_col
        writer_args['id_col'] = args.id_col
    elif writer_type is ARFFWriter:
        writer_args['label_col'] = args.label_col
        writer_args['id_col'] = args.id_col
        writer_args['regression'] = reader.regression
        writer_args['relation'] = reader.relation
    writer = writer_type(args.outfile, feature_set, **writer_args)
    writer.write()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# License: BSD 3 clause
"""
Little helper script to create a summary file out of a list of JSON results
files.

:author: Dan Blanchard (dblanchard@ets.org)
:organization: ETS
"""

import argparse
import logging

from skll.experiments.output import _write_summary_file
from skll.version import __version__


def main(argv=None):
    """
    Handles command line arguments and gets things started.

    Parameters
    ----------
    argv : list of str
        List of arguments, as if specified on the command-line.
        If None, ``sys.argv[1:]`` is used instead.
    """

    # Get command line arguments
    parser = argparse.ArgumentParser(
        description="Creates an experiment summary TSV file from a list of "
                    "JSON files generated by run_experiment.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        conflict_handler='resolve')
    parser.add_argument('summary_file',
                        help='TSV file to store summary of results.')
    parser.add_argument('json_file',
                        help='JSON results file generated by run_experiment.',
                        nargs='+')
    parser.add_argument('-a', '--ablation', action='store_true', default=False,
                        help='The results files are from an ablation run.')
    parser.add_argument('--version', action='version',
                        version=f'%(prog)s {__version__}')
    args = parser.parse_args(argv)

    # Make warnings from built-in warnings module get formatted more nicely
    logging.captureWarnings(True)
    logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - '
                               '%(message)s')

    with open(args.summary_file, 'w') as output_file:
        _write_summary_file(args.json_file, output_file,
                            ablation=int(args.ablation))


if __name__ == '__main__':
    main()



# License: BSD 3 clause
"""
Handles reading and writing data from various types of data files.

:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:author: Nitin Madnani (nmadnani@ets.org)
:organization: ETS
"""

import csv

from .featureset import FeatureSet
from .readers import (
    ARFFReader,
    CSVReader,
    LibSVMReader,
    NDJReader,
    Reader,
    TSVReader,
    safe_float,
)
from .writers import ARFFWriter, CSVWriter, LibSVMWriter, NDJWriter, TSVWriter, Writer

# Register dialect for handling ARFF files
csv.register_dialect('arff', delimiter=',', quotechar="'",
                     escapechar='\\', doublequote=False,
                     lineterminator='\n', skipinitialspace=True)


__all__ = ['Reader', 'safe_float', 'FeatureSet', 'ARFFReader',
           'CSVReader', 'LibSVMReader', 'NDJReader',
           'TSVReader', 'ARFFWriter', 'LibSVMWriter', 'TSVWriter',
           'CSVWriter', 'NDJWriter', 'Writer']



# License: BSD 3 clause
"""
Classes related to storing/merging feature sets.

:author: Dan Blanchard (dblanchard@ets.org)
:author: Nitin Madnani (nmadnani@ets.org)
:author: Jeremy Biggs (jbiggs@ets.org)
:organization: ETS
"""

from copy import deepcopy

import numpy as np
import scipy.sparse as sp
from sklearn.feature_extraction import DictVectorizer, FeatureHasher

from skll.data.dict_vectorizer import DictVectorizer as NewDictVectorizer


class FeatureSet(object):

    """
    Encapsulation of all of the features, values, and metadata about a given
    set of data. This replaces ``ExamplesTuple`` from older versions of SKLL.

    Parameters
    ----------
    name : str
        The name of this feature set.

    ids : np.array of shape (n_ids,)
        Example IDs for this set.

    labels : np.array of shape (n_labels,), default=None
        labels for this set.

    feature : list of dict or an array-like of shape (n_samples, n_features), default=None
        The features for each instance represented as either a
        list of dictionaries or an array-like (if ``vectorizer`` is
        also specified).

    vectorizer : DictVectorizer or FeatureHasher, default=None
        Vectorizer which will be used to generate the feature matrix.

    Warnings
    --------
    FeatureSets can only be equal if the order of the instances is
    identical because these are stored as lists/arrays. Since scikit-learn's
    ``DictVectorizer`` automatically sorts the underlying feature matrix
    if it is sparse, we do not do any sorting before checking for equality.
    This is not a problem because we _always_ use sparse matrices with
    ``DictVectorizer`` when creating FeatureSets.

    Notes
    -----
    If ids, labels, and/or features are not None, the number of rows in
    each array must be equal.
    """

    def __init__(self, name, ids, labels=None, features=None,
                 vectorizer=None):
        super(FeatureSet, self).__init__()
        self.name = name
        if isinstance(ids, list):
            ids = np.array(ids)
        self.ids = ids
        if isinstance(labels, list):
            labels = np.array(labels)
        self.labels = labels
        self.features = features
        self.vectorizer = vectorizer
        # Convert list of dicts to numpy array
        if isinstance(self.features, list):
            if self.vectorizer is None:
                self.vectorizer = NewDictVectorizer(sparse=True)
            self.features = self.vectorizer.fit_transform(self.features)
        if self.features is not None:
            num_feats = self.features.shape[0]
            if self.ids is None:
                raise ValueError('A list of IDs is required')
            num_ids = self.ids.shape[0]
            if num_feats != num_ids:
                raise ValueError(f'Number of IDs ({num_ids}) does not equal '
                                 f'number of feature rows ({num_feats})')
            if self.labels is None:
                self.labels = np.empty(num_feats)
                self.labels.fill(None)
            num_labels = self.labels.shape[0]
            if num_feats != num_labels:
                raise ValueError(f'Number of labels ({num_labels}) does not '
                                 f'equal number of feature rows ({num_feats})')

    def __contains__(self, value):
        """
        Check if example ID is in the FeatureSet.

        Parameters
        ----------
        value
            The value to check.
        """
        return value in self.ids

    def __eq__(self, other):
        """
        Check whether two featuresets are the same.

        Parameters
        ----------
        other : skll.data.FeatureSet
            The other ``FeatureSet`` to check equivalence with.

        Note
        ----
        We consider feature values to be equal if any differences are in the
        sixth decimal place or higher.
        """

        return (self.ids.shape == other.ids.shape and
                self.labels.shape == other.labels.shape and
                self.features.shape == other.features.shape and
                (self.ids == other.ids).all() and
                (self.labels == other.labels).all() and
                np.allclose(self.features.data, other.features.data,
                            rtol=1e-6) and
                (self.features.indices == other.features.indices).all() and
                (self.features.indptr == other.features.indptr).all() and
                self.vectorizer == other.vectorizer)

    def __iter__(self):
        """
        Iterate through (ID, label, feature_dict) tuples in feature set.
        """
        if self.features is not None:
            if not isinstance(self.vectorizer, DictVectorizer):
                raise ValueError('FeatureSets can only be iterated through if '
                                 'they use a DictVectorizer for their feature '
                                 'vectorizer.')
            for id_, label_, feats in zip(self.ids, self.labels, self.features):

                # reshape to a 2D matrix if we are not using a sparse matrix
                # to store the features
                feats = feats.reshape(1, -1) if not sp.issparse(feats) else feats

                # When calling inverse_transform we have to add [0] to get the
                # results for the current instance because it always returns a
                # 2D array
                yield (id_, label_, self.vectorizer.inverse_transform(feats)[0])
        else:
            return

    def __len__(self):
        """
        The number of rows in the ``FeatureSet`` instance.
        """
        return self.features.shape[0]

    def __add__(self, other):
        """
        Combine two feature sets to create a new one.  This is done assuming
        they both have the same instances with the same IDs in the same order.

        Parameters
        ----------
        other : skll.data.FeatureSet
            The other ``FeatureSet`` to add to this one.

        Raises
        ------
        ValueError
            If IDs are not in the same order in each ``FeatureSet`` instance.

        ValueError
            If vectorizers are different between the two ``FeatureSet`` instances.

        ValueError
            If there are duplicate feature names.

        ValueError
            If there are conflicting labels.
        """

        # Check that the sets of IDs are equal
        if set(self.ids) != set(other.ids):
            raise ValueError('IDs are not in the same order in each '
                             'feature set')
        # Compute the relative ordering of IDs for merging the features
        # and labels.
        ids_indices = dict((y, x) for x, y in enumerate(other.ids))
        relative_order = [ids_indices[self_id] for self_id in self.ids]

        # Initialize the new feature set with a name and the IDs.
        new_set = FeatureSet('+'.join(sorted([self.name, other.name])),
                             deepcopy(self.ids))

        # Combine feature matrices and vectorizers.
        if not isinstance(self.vectorizer, type(other.vectorizer)):
            raise ValueError('Cannot combine FeatureSets because they are '
                             'not both using the same type of feature '
                             'vectorizer (e.g., DictVectorizer, '
                             'FeatureHasher)')
        uses_feature_hasher = isinstance(self.vectorizer, FeatureHasher)
        if uses_feature_hasher:
            if (self.vectorizer.n_features !=
                    other.vectorizer.n_features):
                raise ValueError('Cannot combine FeatureSets that uses '
                                 'FeatureHashers with different values of '
                                 'n_features setting.')
        else:
            # Check for duplicate feature names.
            if (set(self.vectorizer.feature_names_) &
                    set(other.vectorizer.feature_names_)):
                raise ValueError('Cannot combine FeatureSets because they '
                                 'have duplicate feature names.')
        num_feats = self.features.shape[1]

        new_set.features = sp.hstack([self.features,
                                      other.features[relative_order]],
                                     'csr')
        new_set.vectorizer = deepcopy(self.vectorizer)
        if not uses_feature_hasher:
            for feat_name, index in other.vectorizer.vocabulary_.items():
                new_set.vectorizer.vocabulary_[feat_name] = (index +
                                                             num_feats)
            other_names = other.vectorizer.feature_names_
            new_set.vectorizer.feature_names_.extend(other_names)

        # If either set has labels, check that they don't conflict.
        if self.has_labels:
            # labels should be the same for each FeatureSet, so store once.
            if other.has_labels and \
                    not np.all(self.labels == other.labels[relative_order]):
                raise ValueError('Feature sets have conflicting labels for '
                                 'examples with the same ID.')
            new_set.labels = deepcopy(self.labels)
        else:
            new_set.labels = deepcopy(other.labels[relative_order])

        return new_set

    def filter(self, ids=None, labels=None, features=None, inverse=False):
        """
        Removes or keeps features and/or examples from the `Featureset` depending
        on the parameters. Filtering is done in-place.

        Parameters
        ----------
        ids : list of str/float, default=None
            Examples to keep in the FeatureSet. If ``None``, no ID
            filtering takes place.

        labels : list of str/float, default=None
            Labels that we want to retain examples for. If ``None``,
            no label filtering takes place.

        features : list of str, default=None
            Features to keep in the FeatureSet. To help with
            filtering string-valued features that were converted
            to sequences of boolean features when read in, any
            features in the FeatureSet that contain a ``=`` will be
            split on the first occurrence and the prefix will be
            checked to see if it is in ``features``.
            If ``None``, no feature filtering takes place.
            Cannot be used if FeatureSet uses a FeatureHasher for
            vectorization.

        inverse : bool, default=False
            Instead of keeping features and/or examples in lists,
            remove them.

        Raises
        ------
        ValueError
            If attempting to use features to filter a ``FeatureSet`` that
            uses a ``FeatureHasher`` vectorizer.
        """
        # Construct mask that indicates which examples to keep
        mask = np.ones(len(self), dtype=bool)
        if ids is not None:
            mask = np.logical_and(mask, np.in1d(self.ids, ids))
        if labels is not None:
            mask = np.logical_and(mask, np.in1d(self.labels, labels))

        if inverse and (labels is not None or ids is not None):
            mask = np.logical_not(mask)

        # Remove examples not in mask
        self.ids = self.ids[mask]
        self.labels = self.labels[mask]
        self.features = self.features[mask, :]

        # Filter features
        if features is not None:
            if isinstance(self.vectorizer, FeatureHasher):
                raise ValueError('FeatureSets with FeatureHasher vectorizers'
                                 ' cannot be filtered by feature.')
            columns = np.array(sorted({feat_num for feat_name, feat_num in
                                       self.vectorizer.vocabulary_.items()
                                       if (feat_name in features or
                                           feat_name.split('=', 1)[0] in
                                           features)}))
            if inverse:
                all_columns = np.arange(self.features.shape[1])
                columns = all_columns[np.logical_not(np.in1d(all_columns,
                                                             columns))]
            self.features = self.features[:, columns]
            self.vectorizer.restrict(columns, indices=True)

    def filtered_iter(self, ids=None, labels=None, features=None,
                      inverse=False):
        """
        A version of `__iter__` that retains only the specified features
        and/or examples from the output.

        Parameters
        ----------
        ids : list of str/float, default=None
            Examples to keep in the ``FeatureSet``. If ``None``, no ID
            filtering takes place.

        labels : list of str/float, default=None
            Labels that we want to retain examples for. If ``None``,
            no label filtering takes place.

        features : list of str, default=None
            Features to keep in the ``FeatureSet``. To help with
            filtering string-valued features that were converted
            to sequences of boolean features when read in, any
            features in the ``FeatureSet`` that contain a `=` will be
            split on the first occurrence and the prefix will be
            checked to see if it is in ``features``.
            If `None`, no feature filtering takes place.
            Cannot be used if ``FeatureSet`` uses a FeatureHasher for
            vectorization.

        inverse : bool, default=False
            Instead of keeping features and/or examples in lists,
            remove them.

        Yields
        ------
        id_ : str
            The ID of the example.

        label_ : str
            The label of the example.

        feat_dict : dict
            The feature dictionary, with feature name as the key
            and example value as the value.

        Raises
        ------
        ValueError
            If the vectorizer is not a ``DictVectorizer``.
        """
        if self.features is not None and not isinstance(self.vectorizer,
                                                        DictVectorizer):
            raise ValueError('FeatureSets can only be iterated through if they'
                             ' use a DictVectorizer for their feature '
                             'vectorizer.')

        for id_, label_, feats in zip(self.ids, self.labels, self.features):
            # Skip instances with IDs not in filter
            if ids is not None and (id_ in ids) == inverse:
                continue
            # Skip instances with labels not in filter
            if labels is not None and (label_ in labels) == inverse:
                continue

            # reshape to a 2D matrix if we are not using a sparse matrix
            # to store the features
            feats = feats.reshape(1, -1) if not sp.issparse(feats) else feats
            feat_dict = self.vectorizer.inverse_transform(feats)[0]
            if features is not None:
                feat_dict = {name: value for name, value in
                             feat_dict.items() if
                             (inverse != (name in features or
                                          name.split('=', 1)[0] in features))}
            elif not inverse:
                feat_dict = {}
            yield id_, label_, feat_dict

    def __sub__(self, other):
        """
        Subset ``FeatureSet`` instance by removing all the features from the
        other ``FeatureSet`` instance.

        Parameters
        ----------
        other : skll.data.FeatureSet
            The other ``FeatureSet`` containing the features that should
            be removed from this ``FeatureSet``.

        Returns
        -------
        A copy of ``self`` with all features in ``other`` removed.
        """
        new_set = deepcopy(self)
        new_set.filter(features=other.vectorizer.feature_names_,
                       inverse=True)
        return new_set

    @property
    def has_labels(self):
        """
        Check if ``FeatureSet`` has finite labels.

        Returns
        -------
        has_labels : bool
            Whether or not this FeatureSet has any finite labels.
        """
        # make sure that labels is not None or a list of Nones
        if self.labels is not None and not all(label is None for label in self.labels):
            # then check that they are not a list of NaNs
            return not (np.issubdtype(self.labels.dtype, np.floating) and
                        np.isnan(np.min(self.labels)))
        else:
            return False

    def __str__(self):
        """
        Returns
        -------
        A string representation of ``FeatureSet``.
        """
        return str(self.__dict__)

    def __repr__(self):
        """
        Returns
        -------
        A string representation of ``FeatureSet``.
        """
        return repr(self.__dict__)

    def __getitem__(self, value):
        """
        Parameters
        ----------
        value
            The value to retrieve.

        Returns
        -------
        A specific example by row number or, if given a slice,
        a new ``FeatureSet`` instance containing a subset of the data.
        """
        # Check if we're slicing
        if isinstance(value, slice):
            sliced_ids = self.ids[value]
            sliced_feats = (self.features[value] if self.features is not None
                            else None)
            sliced_labels = (self.labels[value] if self.labels is not None
                             else None)
            return FeatureSet(f'{self.name}_{value}', sliced_ids,
                              features=sliced_feats, labels=sliced_labels,
                              vectorizer=self.vectorizer)
        else:
            label = self.labels[value] if self.labels is not None else None
            feats = self.features[value, :]
            features = (self.vectorizer.inverse_transform(feats)[0] if
                        self.features is not None else {})
            return self.ids[value], label, features

    @staticmethod
    def split_by_ids(fs, ids_for_split1, ids_for_split2=None):
        """
        Split the ``FeatureSet`` into two new ``FeatureSet`` instances based on
        the given IDs for the two splits.

        Parameters
        ----------
        fs : skll.data.FeatureSet
            The ``FeatureSet`` instance to split.

        ids_for_split1 : list of int
            A list of example IDs which will be split out into
            the first ``FeatureSet`` instance. Note that the
            FeatureSet instance will respect the order of the
            specified IDs.

        ids_for_split2 : list of int, default=None
            An optional ist of example IDs which will be
            split out into the second ``FeatureSet`` instance.
            Note that the ``FeatureSet`` instance will respect
            the order of the specified IDs. If this is
            not specified, then the second ``FeatureSet``
            instance will contain the complement of the
            first set of IDs sorted in ascending order.

        Returns
        -------
        fs1 : skll.data.FeatureSet
            The first ``FeatureSet``.

        fs2 : skll.data.FeatureSet
            The second ``FeatureSet``.
        """

        # Note: an alternative way to implement this is to make copies
        # of the given FeatureSet instance and then use the `filter()`
        # method but that wastes too much memory since it requires making
        # two copies of the original FeatureSet which may be huge. With
        # the current implementation, we are creating new objects but
        # they should be much smaller than the original FeatureSet.
        ids1 = fs.ids[ids_for_split1]
        labels1 = fs.labels[ids_for_split1]
        features1 = fs.features[ids_for_split1]
        if ids_for_split2 is None:
            ids2 = fs.ids[~np.in1d(fs.ids, ids_for_split1)]
            labels2 = fs.labels[~np.in1d(fs.ids, ids_for_split1)]
            features2 = fs.features[~np.in1d(fs.ids, ids_for_split1)]
        else:
            ids2 = fs.ids[ids_for_split2]
            labels2 = fs.labels[ids_for_split2]
            features2 = fs.features[ids_for_split2]

        fs1 = FeatureSet(f'{fs.name}_1',
                         ids1,
                         labels=labels1,
                         features=features1,
                         vectorizer=fs.vectorizer)
        fs2 = FeatureSet(f'{fs.name}_2',
                         ids2,
                         labels=labels2,
                         features=features2,
                         vectorizer=fs.vectorizer)
        return fs1, fs2

    @staticmethod
    def from_data_frame(df, name, labels_column=None, vectorizer=None):
        """
        Helper function to create a ``FeatureSet`` instance from a `pandas.DataFrame`.
        Will raise an Exception if pandas is not installed in your environment.
        The ``ids`` in the ``FeatureSet`` will be the index from the given frame.

        Parameters
        ----------
        df : pd.DataFrame
            The pandas.DataFrame object to use as a ``FeatureSet``.

        name : str
            The name of the output ``FeatureSet`` instance.

        labels_column : str, default=None
            The name of the column containing the labels (data to predict).

        vectorizer : DictVectorizer or FeatureHasher, default=None
            Vectorizer which will be used to generate the feature matrix.

        Returns
        -------
        feature_set : skll.data.FeatureSet
            A ``FeatureSet`` instance generated from from the given data frame.
        """
        if labels_column:
            feature_columns = [column for column in df.columns if column != labels_column]
            labels = df[labels_column].tolist()
        else:
            feature_columns = df.columns
            labels = None

        features = df[feature_columns].to_dict(orient='records')
        return FeatureSet(name,
                          ids=df.index.tolist(),
                          labels=labels,
                          features=features,
                          vectorizer=vectorizer)



#!/usr/bin/env python
# License: BSD 3 clause
from setuptools import find_packages, setup

# Get version without importing, which avoids dependency issues
exec(compile(open('skll/version.py').read(), 'skll/version.py', 'exec'))
# (we use the above instead of execfile for Python 3.x compatibility)


def readme():
    with open('README.rst') as f:
        return f.read()


def requirements():
    req_path = 'requirements.txt'
    with open(req_path) as f:
        reqs = f.read().splitlines()
    return reqs


setup(name='skll',
      version=__version__,  # noqa: F821
      description=('SciKit-Learn Laboratory makes it easier to run machine '
                   'learning experiments with scikit-learn.'),
      long_description=readme(),
      keywords='learning scikit-learn',
      url='http://github.com/EducationalTestingService/skll',
      author='Nitin Madnani',
      author_email='nmadnani@ets.org',
      license='BSD 3 clause',
      packages=find_packages(exclude=['tests', 'examples']),
      entry_points={'console_scripts':
                    ['filter_features = skll.utils.commandline.filter_features:main',
                     'generate_predictions = skll.utils.commandline.generate_predictions:main',
                     'join_features = skll.utils.commandline.join_features:main',
                     'print_model_weights = skll.utils.commandline.print_model_weights:main',
                     'run_experiment = skll.utils.commandline.run_experiment:main',
                     'skll_convert = skll.utils.commandline.skll_convert:main',
                     'summarize_results = skll.utils.commandline.summarize_results:main',
                     'compute_eval_from_predictions = skll.utils.commandline.compute_eval_from_predictions:main',
                     'plot_learning_curves = skll.utils.commandline.plot_learning_curves:main']},
      install_requires=requirements(),
      classifiers=['Intended Audience :: Science/Research',
                   'Intended Audience :: Developers',
                   'License :: OSI Approved :: BSD License',
                   'Programming Language :: Python',
                   'Topic :: Software Development',
                   'Topic :: Scientific/Engineering',
                   'Operating System :: Microsoft :: Windows',
                   'Operating System :: POSIX',
                   'Operating System :: Unix',
                   'Operating System :: MacOS',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.8',
                   'Programming Language :: Python :: 3.9',
                   'Programming Language :: Python :: 3.10',
                   ],
      zip_safe=False)



#!/usr/bin/env python
# License: BSD 3 clause
'''
Script that joins a bunch of feature files together to create one file.

:author: Dan Blanchard (dblanchard@ets.org)
:organization: ETS
'''

import argparse
import logging
import os
import sys

from skll.data.readers import EXT_TO_READER
from skll.data.writers import EXT_TO_WRITER, ARFFWriter, CSVWriter, TSVWriter
from skll.version import __version__


def main(argv=None):
    """
    Handles command line arguments and gets things started.

    Parameters
    ----------
    argv : list of str
        List of arguments, as if specified on the command-line.
        If None, ``sys.argv[1:]`` is used instead.
    """

    # Get command line arguments
    parser = argparse.ArgumentParser(
        description="Joins multiple input feature files together into one "
                    " file.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('infile',
                        help='input feature file (ends in .arff, .csv, '
                             ' .jsonlines, .ndj, or .tsv)',
                        nargs='+')
    parser.add_argument('outfile',
                        help='output feature file')
    parser.add_argument('-i', '--id_col',
                        help='Name of the column which contains the instance '
                             ' IDs in ARFF, CSV, or TSV files.',
                        default='id')
    parser.add_argument('-l', '--label_col',
                        help='Name of the column which contains the class '
                             'labels in ARFF, CSV, or TSV files. For ARFF '
                             'files, this must be the final column to count '
                             'as the label.',
                        default='y')
    parser.add_argument('-q', '--quiet',
                        help='Suppress printing of "Loading..." messages.',
                        action='store_true')
    parser.add_argument('--version', action='version',
                        version=f'%(prog)s {__version__}')
    args = parser.parse_args(argv)

    # Make warnings from built-in warnings module get formatted more nicely
    logging.captureWarnings(True)
    logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - '
                               '%(message)s')
    logger = logging.getLogger(__name__)

    # all extensions except .libsvm can be processed
    valid_extensions = {ext for ext in EXT_TO_READER if ext != '.libsvm'}

    # make sure the input file extensions are those we can process
    input_extension_set = {os.path.splitext(inf)[1].lower() for inf in
                           args.infile}
    output_extension = os.path.splitext(args.outfile)[1].lower()

    # make sure all the files are in the same format except libsvm files
    if len(input_extension_set) > 1:
        logger.error('All input files must be in the same format.')
        sys.exit(1)

    input_extension = list(input_extension_set)[0]

    if input_extension not in valid_extensions:
        logger.error('Input files must be in either .arff, .csv, .jsonlines, '
                     f'.ndj, or .tsv format. You specified: {input_extension}')
        sys.exit(1)

    if output_extension != input_extension:
        logger.error('Output file must be in the same format as the input '
                     f'file.  You specified: {output_extension}')
        sys.exit(1)

    # Read and merge input files
    merged_set = None
    for infile in args.infile:
        reader = EXT_TO_READER[input_extension](infile, quiet=args.quiet,
                                                label_col=args.label_col,
                                                id_col=args.id_col)
        fs = reader.read()
        if merged_set is None:
            merged_set = fs
        else:
            merged_set += fs

    # write out the file in the requested output format
    writer_type = EXT_TO_WRITER[input_extension]
    writer_args = {'quiet': args.quiet}
    if writer_type is CSVWriter or writer_type is TSVWriter:
        writer_args['label_col'] = args.label_col
        writer_args['id_col'] = args.id_col
    elif writer_type is ARFFWriter:
        writer_args['label_col'] = args.label_col
        writer_args['id_col'] = args.id_col
        writer_args['regression'] = reader.regression
        writer_args['relation'] = reader.relation
    writer = writer_type(args.outfile, merged_set, **writer_args)
    writer.write()


if __name__ == '__main__':
    main()
